import asyncio
import logging
import uuid
from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime, timedelta
from enum import Enum
from functools import partial
from numbers import Number
from typing import DefaultDict, Optional, Callable, AsyncGenerator, Any, Iterable, Dict, List, Tuple, TypeVar, cast

from arango import AnalyzerGetError
from arango.collection import VertexCollection, StandardCollection, EdgeCollection
from arango.graph import Graph
from arango.typings import Json
from attr import evolve
from networkx import MultiDiGraph

from resotocore.analytics import CoreEvent, AnalyticsEventSender
from resotocore.core_config import GraphUpdateConfig
from resotocore.db import arango_query, EstimatedSearchCost
from resotocore.db.arango_query import fulltext_delimiter
from resotocore.db.async_arangodb import AsyncArangoDB, AsyncArangoTransactionDB, AsyncArangoDBBase, AsyncCursorContext
from resotocore.db.model import GraphUpdate, QueryModel
from resotocore.db.usagedb import resource_usage_db
from resotocore.error import InvalidBatchUpdate, ConflictingChangeInProgress, NoSuchChangeError, OptimisticLockingFailed
from resotocore.ids import NodeId, GraphName
from resotocore.model.adjust_node import AdjustNode
from resotocore.model.graph_access import GraphAccess, GraphBuilder, EdgeTypes, Section
from resotocore.model.model import (
    Model,
    ComplexKind,
    TransformKind,
    ResolvedProperty,
    UsageDatapoint,
    synthetic_metadata_kinds,
)
from resotocore.model.resolve_in_graph import NodePath, GraphResolver
from resotocore.query.model import Query, FulltextTerm, MergeTerm, P
from resotocore.types import JsonElement, EdgeType
from resotocore.util import first, value_in_path_get, utc_str, uuid_str, value_in_path, json_hash, set_value_in_path

log = logging.getLogger(__name__)


class HistoryChange(Enum):
    node_created = "node_created"
    node_updated = "node_updated"
    node_deleted = "node_deleted"


class GraphDB(ABC):
    @property
    @abstractmethod
    def name(self) -> GraphName:
        pass

    @abstractmethod
    async def get_node(self, model: Model, node_id: NodeId) -> Optional[Json]:
        pass

    @abstractmethod
    async def create_node(self, model: Model, node_id: NodeId, data: Json, under_node_id: NodeId) -> Json:
        pass

    @abstractmethod
    async def update_deferred_edges(self, edges: List[Tuple[NodeId, NodeId, str]], ts: datetime) -> Tuple[int, int]:
        pass

    @abstractmethod
    async def update_node(
        self, model: Model, node_id: NodeId, patch_or_replace: Json, replace: bool, section: Optional[str]
    ) -> Json:
        pass

    @abstractmethod
    def update_nodes(
        self, model: Model, patches_by_id: Dict[NodeId, Json], **kwargs: Any
    ) -> AsyncGenerator[Json, None]:
        pass

    @abstractmethod
    def update_nodes_desired(
        self, model: Model, patch: Json, node_ids: List[NodeId], **kwargs: Any
    ) -> AsyncGenerator[Json, None]:
        pass

    @abstractmethod
    def update_nodes_metadata(
        self, model: Model, patch: Json, node_ids: List[NodeId], **kwargs: Any
    ) -> AsyncGenerator[Json, None]:
        pass

    @abstractmethod
    async def delete_node(self, node_id: NodeId) -> None:
        pass

    @abstractmethod
    async def merge_graph(
        self,
        graph_to_merge: MultiDiGraph,
        model: Model,
        maybe_change_id: Optional[str] = None,
        is_batch: bool = False,
    ) -> Tuple[List[str], GraphUpdate]:
        pass

    @abstractmethod
    async def list_in_progress_updates(self) -> List[Json]:
        pass

    @abstractmethod
    async def commit_batch_update(self, batch_id: str) -> None:
        pass

    @abstractmethod
    async def abort_update(self, batch_id: str) -> None:
        pass

    @abstractmethod
    async def search_list(
        self, query: QueryModel, with_count: bool = False, timeout: Optional[timedelta] = None, **kwargs: Any
    ) -> AsyncCursorContext:
        pass

    @abstractmethod
    async def search_history(
        self,
        query: QueryModel,
        change: Optional[HistoryChange] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        with_count: bool = False,
        timeout: Optional[timedelta] = None,
        **kwargs: Any,
    ) -> AsyncCursorContext:
        pass

    @abstractmethod
    async def search_graph_gen(
        self, query: QueryModel, with_count: bool = False, timeout: Optional[timedelta] = None
    ) -> AsyncCursorContext:
        pass

    @abstractmethod
    async def search_graph(self, query: QueryModel) -> MultiDiGraph:
        pass

    @abstractmethod
    async def search_aggregation(self, query: QueryModel) -> AsyncCursorContext:
        pass

    @abstractmethod
    async def explain(self, query: QueryModel, with_edges: bool = False) -> EstimatedSearchCost:
        pass

    @abstractmethod
    async def wipe(self) -> None:
        pass

    @abstractmethod
    async def to_query(self, query_model: QueryModel, with_edges: bool = False) -> Tuple[str, Json]:
        pass

    @abstractmethod
    async def create_update_schema(self) -> None:
        pass

    @abstractmethod
    async def copy_graph(self, to_graph: GraphName, to_snapshot: bool = False) -> "GraphDB":
        pass

    @abstractmethod
    async def insert_usage_data(self, data: List[UsageDatapoint]) -> None:
        pass


class ArangoGraphDB(GraphDB):
    def __init__(self, db: AsyncArangoDB, name: GraphName, adjust_node: AdjustNode, config: GraphUpdateConfig) -> None:
        super().__init__()
        self._name = name
        self.node_adjuster = adjust_node
        self.vertex_name = name
        self.in_progress = f"{name}_in_progress"
        self.node_history = f"{name}_node_history"
        self.usage_db = resource_usage_db(db, f"{name}_usage")
        self.db = db
        self.config = config

    @property
    def name(self) -> GraphName:
        return self._name

    def edge_collection(self, edge_type: EdgeType) -> str:
        return f"{self.name}_{edge_type}"

    async def get_node(self, model: Model, node_id: NodeId) -> Optional[Json]:
        node = await self.by_id(node_id)
        return self.document_to_instance_fn(model)(node) if node is not None else None

    async def create_node(self, model: Model, node_id: NodeId, data: Json, under_node_id: NodeId) -> Json:
        graph = GraphBuilder(model, uuid_str())
        graph.add_node(node_id, data)
        graph.add_edge(under_node_id, node_id, EdgeTypes.default)
        access = GraphAccess(graph.graph, node_id, {under_node_id})
        _, node_inserts, _, _ = self.prepare_nodes(access, [], model)
        _, edge_inserts, _ = self.prepare_edges(access, [], EdgeTypes.default)
        assert len(node_inserts) == 1
        assert len(edge_inserts) == 1
        edge_collection = self.edge_collection(EdgeTypes.default)
        async with self.db.begin_transaction(write=[self.vertex_name, edge_collection]) as tx:
            result: Json = await tx.insert(self.vertex_name, node_inserts[0], return_new=True)  # type: ignore
            await tx.insert(edge_collection, edge_inserts[0])
            trafo = self.document_to_instance_fn(model)
            return trafo(result["new"])

    async def update_deferred_edges(self, edges: List[Tuple[NodeId, NodeId, str]], ts: datetime) -> Tuple[int, int]:
        default_edges: List[Json] = []
        delete_edges: List[Json] = []

        for from_node, to_node, edge_type in edges:
            json_node = self.edge_to_json(from_node, to_node, None)
            json_node["outer_edge_ts"] = utc_str(ts)  # must be kept in sync with an index
            if edge_type == EdgeTypes.default:
                default_edges.append(json_node)
            else:
                delete_edges.append(json_node)

        def deletion_query(edge_collection: str) -> str:
            return f"""
            FOR edge IN {edge_collection}
                FILTER edge.outer_edge_ts != null &&  edge.outer_edge_ts < "{utc_str(ts)}"
                REMOVE edge in {edge_collection}
            """

        updated_edges = 0
        deleted_edges = 0

        if default_edges:
            edge_collection = self.edge_collection(EdgeTypes.default)
            async with self.db.begin_transaction(write=[edge_collection]) as tx:
                await tx.insert_many(edge_collection, default_edges, overwrite=True)
                updated_edges += len(default_edges)
                query = deletion_query(edge_collection)
                with await tx.aql(query, count=True) as cursor:
                    deleted_edges += cursor.count() or 0

        if delete_edges:
            edge_collection = self.edge_collection(EdgeTypes.delete)
            async with self.db.begin_transaction(write=[edge_collection]) as tx:
                await tx.insert_many(edge_collection, delete_edges, overwrite=True)
                updated_edges += len(delete_edges)
                query = deletion_query(edge_collection)
                with await tx.aql(query, count=True) as cursor:
                    deleted_edges += cursor.count() or 0

        return updated_edges, deleted_edges

    async def update_node(
        self, model: Model, node_id: NodeId, patch_or_replace: Json, replace: bool, section: Optional[str]
    ) -> Json:
        return await self.update_node_with(self.db, model, node_id, patch_or_replace, replace, section)

    async def update_node_with(
        self,
        db: AsyncArangoDBBase,
        model: Model,
        node_id: NodeId,
        patch_or_replace: Json,
        replace: bool,
        section: Optional[str],
    ) -> Json:
        node = await self.by_id_with(db, node_id)
        if node is None:
            raise AttributeError(f"No document found with this id: {node_id}")
        if "revision" in patch_or_replace and patch_or_replace["revision"] != node["_rev"]:
            raise OptimisticLockingFailed(node_id)

        updated = node.copy()
        if section:
            existing_section = node.get(section)
            existing_section = existing_section if existing_section else {}
            updated[section] = patch_or_replace if replace else {**existing_section, **patch_or_replace}
        else:
            for sect in Section.content_ordered:
                if sect in patch_or_replace:
                    existing_section = node.get(sect)
                    existing_section = existing_section if existing_section else {}
                    updated[sect] = (
                        patch_or_replace[sect] if replace else {**existing_section, **patch_or_replace[sect]}
                    )

        # Only the reported section is defined by the model and can be coerced
        kind = model[updated[Section.reported]]
        coerced = kind.check_valid(updated[Section.reported])
        updated[Section.reported] = coerced if coerced is not None else updated[Section.reported]

        # call adjuster on resulting node
        ctime = value_in_path_get(node, NodePath.reported_ctime, utc_str())
        adjusted = self.adjust_node(model, GraphAccess.dump_direct(node_id, updated, kind, recompute=True), ctime)
        update = {"_key": node["_key"], "hash": adjusted["hash"], "kinds": adjusted["kinds"], "flat": adjusted["flat"]}
        # copy relevant sections into update node
        for sec in [section] if section else Section.content_ordered:
            if sec in adjusted:
                update[sec] = adjusted[sec]

        result = await db.update(self.vertex_name, update, return_new=True, merge=not replace)
        trafo = self.document_to_instance_fn(model)
        return trafo(result["new"])

    async def update_nodes(
        self, model: Model, patches_by_id: Dict[NodeId, Json], **kwargs: Any
    ) -> AsyncGenerator[Json, None]:
        log.info(f"Update nodes called with {len(patches_by_id)} updates.")
        # collect all sections to be deleted
        deletes: Dict[str, List[str]] = defaultdict(list)
        delete_sections = [Section.desired, Section.metadata]
        # group patches by changes: single desired or metadata changes can be executed via special purpose methods.
        updates: Dict[str, Json] = {}
        updated_nodes: Dict[str, List[NodeId]] = defaultdict(list)
        for uid, patch_js in patches_by_id.items():
            # filter out delete operation
            for section in delete_sections:
                if section in patch_js and patch_js[section] is None:
                    deletes[section].append(uid)
                    del patch_js[section]
            # filter out empty changes (== noop patches)
            for section in Section.content_ordered:
                if section in patch_js and patch_js[section] == {}:
                    del patch_js[section]
            # all remaining changes are updates
            if patch_js:
                hashed = json_hash(patch_js)
                updates[hashed] = patch_js
                updated_nodes[hashed].append(uid)

        # all changes are executed inside a transaction: either all changes are successful or none
        async with self.db.begin_transaction(read=[self.vertex_name], write=[self.vertex_name]) as tx:

            async def update_node_multi(js: Json, node_ids: List[NodeId]) -> AsyncGenerator[Json, None]:
                for node_id in node_ids:
                    log.debug(f"Update node: change={js} on {node_id}")
                    single_update = await self.update_node_with(tx, model, node_id, js, False, None)
                    yield single_update

            for section, ids in deletes.items():
                log.debug(f"Delete section {section} for ids: {ids}")
                async for res in self.delete_nodes_section_with(tx, model, section, ids):
                    yield res

            for change_id, change in updates.items():
                items = updated_nodes[change_id]
                if len(change) == 1 and Section.desired in change:
                    log.debug(f"Update desired many: change={change} on {items}")
                    patch = change[Section.desired]
                    result = self.update_nodes_section_with(tx, model, Section.desired, patch, items)
                elif len(change) == 1 and Section.metadata in change:
                    log.debug(f"Update metadata many: change={change} on {items}")
                    patch = change[Section.metadata]
                    result = self.update_nodes_section_with(tx, model, Section.metadata, patch, items)
                else:
                    result = update_node_multi(change, items)
                async for res in result:
                    yield res

    def update_nodes_desired(
        self, model: Model, patch: Json, node_ids: List[NodeId], **kwargs: Any
    ) -> AsyncGenerator[Json, None]:
        return self.update_nodes_section_with(self.db, model, Section.desired, patch, node_ids)

    def update_nodes_metadata(
        self, model: Model, patch: Json, node_ids: List[NodeId], **kwargs: Any
    ) -> AsyncGenerator[Json, None]:
        return self.update_nodes_section_with(self.db, model, Section.metadata, patch, node_ids)

    async def delete_nodes_section_with(
        self, db: AsyncArangoDBBase, model: Model, section: str, node_ids: List[str]
    ) -> AsyncGenerator[Json, None]:
        bind_var = {"node_ids": node_ids}
        trafo = self.document_to_instance_fn(model)
        with await db.aql(query=self.query_delete_desired_metadata_many(section), bind_vars=bind_var) as cursor:
            for element in cursor:
                yield trafo(element)

    async def update_nodes_section_with(
        self, db: AsyncArangoDBBase, model: Model, section: str, patch: Json, node_ids: List[NodeId]
    ) -> AsyncGenerator[Json, None]:
        bind_var = {"patch": patch, "node_ids": node_ids}
        trafo = self.document_to_instance_fn(model)
        with await db.aql(query=self.query_update_desired_metadata_many(section), bind_vars=bind_var) as cursor:
            for element in cursor:
                yield trafo(element)

    async def delete_node(self, node_id: NodeId) -> None:
        with await self.db.aql(query=self.query_count_direct_children(), bind_vars={"rid": node_id}) as cursor:
            count = cursor.next()
            if count > 0:
                raise AttributeError(f"Can not delete node, since it has {count} child(ren)!")

        with await self.db.aql(query=self.query_node_by_id(), bind_vars={"rid": node_id}) as cursor:
            if not cursor.empty():
                await self.db.delete_vertex(self.name, cursor.next())
            else:
                return None

    async def by_id(self, node_id: NodeId) -> Optional[Json]:
        return await self.by_id_with(self.db, node_id)

    async def by_id_with(self, db: AsyncArangoDBBase, node_id: NodeId) -> Optional[Json]:
        with await db.aql(query=self.query_node_by_id(), bind_vars={"rid": node_id}) as cursor:
            return cursor.next() if not cursor.empty() else None

    async def search_list(
        self, query: QueryModel, with_count: bool = False, timeout: Optional[timedelta] = None, **kwargs: Any
    ) -> AsyncCursorContext:
        assert query.query.aggregate is None, "Given query is an aggregation function. Use the appropriate endpoint!"
        q_string, bind = await self.to_query(query)
        return await self.db.aql_cursor(
            query=q_string,
            trafo=self.document_to_instance_fn(query.model, query.query),
            count=with_count,
            bind_vars=bind,
            batch_size=10000,
            ttl=cast(Number, int(timeout.total_seconds())) if timeout else None,
        )

    async def search_history(
        self,
        query: QueryModel,
        change: Optional[HistoryChange] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        with_count: bool = False,
        timeout: Optional[timedelta] = None,
        **kwargs: Any,
    ) -> AsyncCursorContext:
        more_than_one = len(query.query.parts) > 1
        has_invalid_terms = any(query.query.find_terms(lambda t: isinstance(t, (FulltextTerm, MergeTerm))))
        has_navigation = any(p.navigation for p in query.query.parts)
        if more_than_one and has_invalid_terms or has_navigation:
            raise AttributeError("Fulltext, merge terms and navigation is not supported in history queries!")
        # adjust query
        term = query.query.current_part.term
        if change:
            term = term.and_term(P.single("change").eq(change.value))
        if after:
            term = term.and_term(P.single("changed_at").gt(utc_str(after)))
        if before:
            term = term.and_term(P.single("changed_at").lt(utc_str(before)))
        query = QueryModel(evolve(query.query, parts=[evolve(query.query.current_part, term=term)]), query.model)
        q_string, bind = arango_query.to_query(self, query, from_collection=self.node_history)
        trafo = (
            None
            if query.query.aggregate
            else self.document_to_instance_fn(
                query.model, query.query, ["change", "changed_at", "created", "updated", "deleted"]
            )
        )
        ttl = cast(Number, int(timeout.total_seconds())) if timeout else None
        return await self.db.aql_cursor(
            query=q_string, trafo=trafo, count=with_count, bind_vars=bind, batch_size=10000, ttl=ttl
        )

    async def search_graph_gen(
        self, query: QueryModel, with_count: bool = False, timeout: Optional[timedelta] = None
    ) -> AsyncCursorContext:
        assert query.query.aggregate is None, "Given query is an aggregation function. Use the appropriate endpoint!"
        query_string, bind = await self.to_query(query, with_edges=True)
        return await self.db.aql_cursor(
            query=query_string,
            trafo=self.document_to_instance_fn(query.model, query.query),
            bind_vars=bind,
            count=with_count,
            batch_size=10000,
            ttl=cast(Number, int(timeout.total_seconds())) if timeout else None,
        )

    async def search_graph(self, query: QueryModel) -> MultiDiGraph:
        async with await self.search_graph_gen(query) as cursor:
            graph = MultiDiGraph()
            async for item in cursor:
                if "from" in item and "to" in item and "edge_type" in item:
                    key = GraphAccess.edge_key(item["from"], item["to"], item["edge_type"])
                    graph.add_edge(key.from_node, key.to_node, key, edge_type=key.edge_type)
                elif "id" in item:
                    graph.add_node(item["id"], **item)
            return graph

    async def search_aggregation(self, query: QueryModel) -> AsyncCursorContext:
        q_string, bind = await self.to_query(query)
        assert query.query.aggregate is not None, "Given query has no aggregation section"
        return await self.db.aql_cursor(query=q_string, bind_vars=bind)

    async def explain(self, query: QueryModel, with_edges: bool = False) -> EstimatedSearchCost:
        return await arango_query.query_cost(self, query, with_edges)

    async def wipe(self) -> None:
        await self.db.truncate(self.vertex_name)
        for edge_type in EdgeTypes.all:
            await self.db.truncate(self.edge_collection(edge_type))
        await self.insert_genesis_data()

    @staticmethod
    def document_to_instance_fn(
        model: Model, query: Optional[Query] = None, additional_root_props: Optional[List[str]] = None
    ) -> Callable[[Json], Json]:
        synthetic_metadata = model.predefined_synthetic_props(synthetic_metadata_kinds)

        def props(doc: Json, result: Json, definition: Iterable[str]) -> None:
            for prop in definition:
                if prop in doc and doc[prop]:
                    result[prop] = doc[prop]

        def synth_props(doc: Json, result: Json, section: str, synthetic_properties: List[ResolvedProperty]) -> None:
            if (section_in := doc.get(section)) and (section_out := result.get(section)):
                for synth in synthetic_properties:
                    if isinstance(synth.kind, TransformKind) and synth.prop.synthetic:
                        source_value = value_in_path(section_in, synth.prop.synthetic.path)
                        if source_value:
                            section_out[synth.prop.name] = synth.kind.transform(source_value)

        def render_prop(doc: Json, root_level: bool) -> Json:
            if Section.reported in doc or Section.desired in doc or Section.metadata in doc:
                # side note: the dictionary remembers insertion order
                # this order is also used to render the output (e.g. yaml property order)
                result = {"id": doc["_key"], "type": "node"}
                if "_rev" in doc:
                    result["revision"] = doc["_rev"]
                props(doc, result, Section.content)
                if root_level:
                    props(doc, result, Section.lookup_sections_ordered)
                    if additional_root_props:
                        props(doc, result, additional_root_props)
                kind = model.get(doc[Section.reported])
                if isinstance(kind, ComplexKind):
                    synth_props(doc, result, Section.reported, kind.synthetic_props())
                    synth_props(doc, result, Section.metadata, synthetic_metadata)
                return result
            else:
                return doc

        def render_merge_results(doc: Json, result: Json, q: Query) -> Json:
            for mq in q.merge_query_by_name:
                merged = value_in_path(doc, mq.name)
                if merged:
                    if mq.only_first and isinstance(merged, dict):
                        rd = render_merge_results(merged, render_prop(merged, False), mq.query)
                        set_value_in_path(rd, mq.name, result)
                    elif isinstance(merged, list):
                        rl = [render_merge_results(elem, render_prop(elem, False), mq.query) for elem in merged]
                        set_value_in_path(rl, mq.name, result)
            return result

        def merge_results(doc: Json) -> Json:
            rendered = render_prop(doc, True)
            if query:
                render_merge_results(doc, rendered, query)
            return rendered

        return merge_results

    async def list_in_progress_updates(self) -> List[Json]:
        with await self.db.aql(self.query_active_updates()) as cursor:
            return list(cursor)

    async def get_tmp_collection(self, change_id: str, create: bool = True) -> StandardCollection:
        id_part = str(uuid.uuid5(uuid.NAMESPACE_DNS, change_id)).replace("-", "")
        temp_name = f"{self.vertex_name}_temp_{id_part}"
        if await self.db.has_collection(temp_name):
            return self.db.collection(temp_name)
        elif create:
            temp = await self.db.create_collection(temp_name, replication_factor=1)
            temp.add_persistent_index(["action"])
            return temp
        else:
            raise NoSuchChangeError(change_id)

    async def move_temp_to_proper(self, change_id: str, temp_name: str) -> None:
        change_key = str(uuid.uuid5(uuid.NAMESPACE_DNS, change_id))
        log.info(f"Move temp->proper data: change_id={change_id}, change_key={change_key}, temp_name={temp_name}")
        edge_inserts = [
            f'for e in {temp_name} filter e.action=="edge_insert" and e.edge_type=="{a}" '
            f'insert e.data in {self.edge_collection(a)} OPTIONS {{overwriteMode: "replace"}}'
            for a in EdgeTypes.all
        ]
        edge_deletes = [
            f'for e in {temp_name} filter e.action=="edge_delete" and e.edge_type=="{a}" '
            f"remove e.data in {self.edge_collection(a)}"
            for a in EdgeTypes.all
        ]
        history_updates = [
            f'for e in {temp_name} filter e.action=="node_created" insert MERGE({{change: e.action, changed_at: e.data.created}}, UNSET(e.data, "_key", "flat", "hash")) in {self.node_history}',  # noqa: E501
            f'for e in {temp_name} filter e.action=="node_updated" insert MERGE({{change: e.action, changed_at: e.data.updated}}, UNSET(e.data, "_key", "flat", "hash")) in {self.node_history}',  # noqa: E501
            f'for e in {temp_name} filter e.action=="node_deleted" let node = Document(CONCAT("{self.vertex_name}/", e.data._key)) insert MERGE({{change: "node_deleted", deleted: e.data.deleted, changed_at: e.data.deleted}}, UNSET(node, "_key", "_id", "_rev", "flat", "hash")) in {self.node_history}',  # noqa: E501
        ]
        usage_updates = [
            f'for u in {self.usage_db.collection_name} filter u.change_id=="{change_id}" update {{_key: u.id}} with {{ usage: MERGE(u.v, {{ start: DATE_ISO8601(u.at*1000), duration: "1h" }}) }} in {self.vertex_name} options {{ mergeObjects: false }}'  # noqa: E501
        ]
        updates = ";\n".join(
            map(
                lambda aql: f"db._createStatement({{ query: `{aql}` }}).execute()",
                (history_updates if self.config.keep_history else [])
                + [
                    f'for e in {temp_name} filter e.action=="node_created" insert e.data in {self.vertex_name}'
                    ' OPTIONS{overwriteMode: "replace"}',
                    f'for e in {temp_name} filter e.action=="node_updated" update e.data in {self.vertex_name}'
                    " OPTIONS {mergeObjects: false}",
                    f'for e in {temp_name} filter e.action=="node_deleted" remove e.data in {self.vertex_name}',
                ]
                + edge_inserts
                + edge_deletes
                + usage_updates
                + [f'remove {{_key: "{change_key}"}} in {self.in_progress}'],
            )
        )
        await self.db.execute_transaction(
            f'function () {{\nvar db=require("@arangodb").db;\n{updates}\n}}',
            read=[temp_name, self.usage_db.collection_name],
            write=[self.edge_collection(a) for a in EdgeTypes.all]
            + [self.vertex_name, self.in_progress, self.node_history],
        )
        log.info(f"Move temp->proper data: change_id={change_id} done.")

    async def mark_update(
        self, root_node_ids: List[str], parent_node_ids: List[str], change_id: str, is_batch: bool
    ) -> None:
        async with self.db.begin_transaction(read=[self.in_progress], write=[self.in_progress]) as tx:
            existing = next(await tx.aql(self.query_active_change(), bind_vars={"root_node_ids": root_node_ids}), None)
            if existing is not None:
                other = existing["change"]
                raise InvalidBatchUpdate() if change_id == other else ConflictingChangeInProgress(other)
            await tx.insert(
                self.in_progress,
                {
                    "_key": str(uuid.uuid5(uuid.NAMESPACE_DNS, change_id)),
                    "root_node_ids": list(root_node_ids),
                    "parent_node_ids": list(parent_node_ids),
                    "change": change_id,
                    "created": utc_str(),
                    "is_batch": is_batch,
                },
            )

    async def refresh_marked_update(self, change_id: str) -> None:
        with await self.db.aql(self.update_active_change(), bind_vars={"change": change_id}):
            return None

    async def delete_marked_update(self, change_id: str, tx: Optional[AsyncArangoTransactionDB] = None) -> None:
        db = tx if tx else self.db
        doc = {"_key": str(uuid.uuid5(uuid.NAMESPACE_DNS, change_id))}
        await db.delete(self.in_progress, doc, ignore_missing=True)

    def adjust_node(self, model: Model, json: Json, created_at: Any) -> Json:
        reported = json[Section.reported]
        # preserve ctime in reported: if it is not set, use the creation time of the object
        if not reported.get("ctime", None):
            kind = model[reported]
            if isinstance(kind, ComplexKind) and "ctime" in kind:
                reported["ctime"] = created_at

        # adjuster has the option to manipulate the resulting json
        return self.node_adjuster.adjust(json)

    def prepare_nodes(
        self, access: GraphAccess, node_cursor: Iterable[Json], model: Model
    ) -> Tuple[GraphUpdate, List[Json], List[Json], List[Json]]:
        log.info(f"Prepare nodes for subgraph {access.root()}")
        info = GraphUpdate()
        resource_inserts: List[Json] = []
        resource_updates: List[Json] = []
        resource_deletes: List[Json] = []

        optional_properties = [*Section.all_ordered, "refs", "kinds", "flat", "hash"]

        def insert_node(node: Json) -> None:
            elem = self.adjust_node(model, node, access.at_json)
            js_doc: Json = {"_key": elem["id"], "created": access.at_json, "updated": access.at_json}
            for prop in optional_properties:
                value = node.get(prop, None)
                if value:
                    js_doc[prop] = value
            resource_inserts.append(js_doc)
            info.nodes_created += 1

        def update_or_delete_node(node: Json) -> None:
            key = node["_key"]
            hash_string = node["hash"]
            elem = access.node(key)
            if elem is None:
                # node is in db, but not in the graph any longer: delete node
                resource_deletes.append({"_key": key, "deleted": access.at_json})
                info.nodes_deleted += 1
            elif elem["hash"] != hash_string:
                # node is in db and in the graph, content is different
                adjusted: Json = self.adjust_node(model, elem, node["created"])
                js = {"_key": key, "created": node["created"], "updated": access.at_json}
                for prop in optional_properties:
                    value = adjusted.get(prop, None)
                    if value:
                        js[prop] = value
                resource_updates.append(js)
                info.nodes_updated += 1

        for doc in node_cursor:
            update_or_delete_node(doc)

        for not_visited in access.not_visited_nodes():
            insert_node(not_visited)
        return info, resource_inserts, resource_updates, resource_deletes

    def edge_to_json(self, from_node: str, to_node: str, refs: Optional[Dict[str, str]]) -> Json:
        key = self.db_edge_key(from_node, to_node)
        js = {
            "_key": key,
            "_from": f"{self.vertex_name}/{from_node}",
            "_to": f"{self.vertex_name}/{to_node}",
            "refs": refs,
        }
        return js

    def prepare_edges(
        self, access: GraphAccess, edge_cursor: Iterable[Json], edge_type: EdgeType
    ) -> Tuple[GraphUpdate, List[Json], List[Json]]:
        log.info(f"Prepare edges of type {edge_type} for subgraph {access.root()}")
        info = GraphUpdate()
        edges_inserts: List[Json] = []
        edges_deletes: List[Json] = []

        def insert_edge(from_node: str, to_node: str) -> None:
            # Take the refs with the lower number of entries (or none):
            # Lower number of entries means closer to the root.
            # Ownership is maintained as self-contained subgraph.
            # A relationship from a sub-graph root to a node closer to the graph root, is not considered
            # part of the sub-graph root, but the parent graph.
            to_refs = access.nodes[to_node].get("refs")
            from_refs = access.nodes[from_node].get("refs")
            refs = (to_refs if len(to_refs) < len(from_refs) else from_refs) if to_refs and from_refs else None

            js = self.edge_to_json(from_node, to_node, refs)
            edges_inserts.append(js)
            info.edges_created += 1

        def update_edge(edge: Json) -> None:
            from_node = edge["_from"].split("/")[1]  # vertex/id
            to_node = edge["_to"].split("/")[1]  # vertex/id
            if not access.has_edge(from_node, to_node, edge_type):
                edges_deletes.append(edge)
                info.edges_deleted += 1

        for doc in edge_cursor:
            update_edge(doc)

        for edge_from, edge_to in access.not_visited_edges(edge_type):
            insert_edge(edge_from, edge_to)

        return info, edges_inserts, edges_deletes

    async def merge_graph(
        self,
        graph_to_merge: MultiDiGraph,
        model: Model,
        maybe_change_id: Optional[str] = None,
        is_batch: bool = False,
    ) -> Tuple[List[str], GraphUpdate]:
        change_id = maybe_change_id if maybe_change_id else uuid_str()

        async def prepare_graph(
            sub: GraphAccess, node_query: Tuple[str, Json], edge_query: Callable[[EdgeType], Tuple[str, Json]]
        ) -> Tuple[
            GraphUpdate, List[Json], List[Json], List[Json], Dict[EdgeType, List[Json]], Dict[EdgeType, List[Json]]
        ]:
            graph_info = GraphUpdate()
            # check all nodes for this subgraph
            query, bind = node_query
            log.debug(f"Query for nodes: {sub.root()}")
            with await self.db.aql(query, bind_vars=bind, batch_size=50000) as node_cursor:
                node_info, ni, nu, nd = self.prepare_nodes(sub, node_cursor, model)
                graph_info += node_info

            # check all edges in all relevant edge-collections
            edge_inserts: DefaultDict[EdgeType, List[Json]] = defaultdict(list)
            edge_deletes: DefaultDict[EdgeType, List[Json]] = defaultdict(list)
            for edge_type in EdgeTypes.all:
                query, bind = edge_query(edge_type)
                log.debug(f"Query for edges of type {edge_type}: {sub.root()}")
                with await self.db.aql(query, bind_vars=bind, batch_size=50000) as ec:
                    edge_info, gei, ged = self.prepare_edges(sub, ec, edge_type)
                    graph_info += edge_info
                    edge_inserts[edge_type] = gei
                    edge_deletes[edge_type] = ged
            return graph_info, ni, nu, nd, edge_inserts, edge_deletes

        roots, parent, graphs = GraphAccess.merge_graphs(graph_to_merge)
        logging.info(f"merge_graph {len(roots)} merge nodes found. change_id={change_id}, is_batch={is_batch}.")

        def parent_edges(edge_type: EdgeType) -> Tuple[str, Json]:
            edge_ids = [self.db_edge_key(f, t) for f, t, et in parent.g.edges(data="edge_type") if et == edge_type]
            return self.query_update_edges_by_ids(edge_type), {"ids": edge_ids}

        def merge_edges(merge_node: str, merge_node_kind: str, edge_type: EdgeType) -> Tuple[str, Json]:
            return self.query_update_edges(edge_type, merge_node_kind), {"update_id": merge_node}

        K = TypeVar("K")  # noqa: N806
        V = TypeVar("V")  # noqa: N806

        def combine_dict(left: Dict[K, List[V]], right: Dict[K, List[V]]) -> Dict[K, List[V]]:
            result = dict(left)
            for key, right_values in right.items():
                left_values = left.get(key)
                result[key] = left_values + right_values if left_values else right_values
            return result

        # this will throw an exception, in case of a conflicting update (--> outside try block)
        log.debug("Mark all parent nodes for this update to avoid conflicting changes")
        await self.mark_update(roots, list(parent.nodes), change_id, is_batch)
        try:
            parents_nodes = self.query_update_nodes_by_ids(), {"ids": list(parent.g.nodes)}
            info, nis, nus, nds, eis, eds = await prepare_graph(parent, parents_nodes, parent_edges)
            for num, (root, graph) in enumerate(graphs):
                root_kind = GraphResolver.resolved_kind(graph_to_merge.nodes[root])
                if root_kind:
                    log.info(f"Update subgraph: root={root} ({root_kind}, {num+1} of {len(roots)})")
                    node_query = self.query_update_nodes(root_kind), {"update_id": root}
                    edge_query = partial(merge_edges, root, root_kind)

                    i, ni, nu, nd, ei, ed = await prepare_graph(graph, node_query, edge_query)
                    info += i
                    nis += ni
                    nus += nu
                    nds += nd
                    eis = combine_dict(eis, ei)
                    eds = combine_dict(eds, ed)
                else:
                    # Already checked in GraphAccess - only here as safeguard.
                    raise AttributeError(f"Kind of update root {root} is not a pre-resolved and can not be used!")

            log.debug(f"Update prepared: {info}. Going to persist the changes.")
            await self.refresh_marked_update(change_id)
            await self.persist_update(change_id, is_batch, info, nis, nus, nds, eis, eds)
            return roots, info
        except Exception as ex:
            await self.delete_marked_update(change_id)
            raise ex

    async def persist_update(
        self,
        change_id: str,
        is_batch: bool,
        info: GraphUpdate,
        resource_inserts: List[Json],
        resource_updates: List[Json],
        resource_deletes: List[Json],
        edge_inserts: Dict[EdgeType, List[Json]],
        edge_deletes: Dict[EdgeType, List[Json]],
    ) -> None:
        async def execute_many_async(
            async_fn: Callable[[str, List[Json]], Any], name: str, array: List[Json], **kwargs: Any
        ) -> None:
            if array:
                async_fn_with_args = partial(async_fn, **kwargs) if kwargs else async_fn
                result = await async_fn_with_args(name, array)  # type: ignore
                ex: Optional[Exception] = first(lambda x: isinstance(x, Exception), result)
                if ex:
                    raise ex  # pylint: disable=raising-bad-type

        async def trafo_many(
            async_fn: Callable[[str, List[Json]], Any], name: str, array: List[Json], template: Json
        ) -> None:
            # update the array in place to not create another intermediate array
            for idx, item in enumerate(array):
                entry = template.copy()
                entry["data"] = item
                array[idx] = entry
            await execute_many_async(async_fn, name, array)

        async def store_to_tmp_collection(temp: StandardCollection) -> None:
            tmp = temp.name
            ri = trafo_many(self.db.insert_many, tmp, resource_inserts, {"action": "node_created"})
            ru = trafo_many(self.db.insert_many, tmp, resource_updates, {"action": "node_updated"})
            rd = trafo_many(self.db.insert_many, tmp, resource_deletes, {"action": "node_deleted"})
            edge_i = [
                trafo_many(self.db.insert_many, tmp, inserts, {"action": "edge_insert", "edge_type": tpe})
                for tpe, inserts in edge_inserts.items()
            ]
            edge_u = [
                trafo_many(self.db.insert_many, tmp, deletes, {"action": "edge_delete", "edge_type": tpe})
                for tpe, deletes in edge_deletes.items()
            ]
            await asyncio.gather(*([ri, ru, rd] + edge_i + edge_u))

        async def update_via_temp_collection() -> None:
            temp = await self.get_tmp_collection(change_id)
            log.debug(f"Store change in temp collection {temp.name}")
            try:
                await store_to_tmp_collection(temp)
                await self.move_temp_to_proper(change_id, temp.name)
            finally:
                log.debug(f"Delete temp collection {temp.name}")
                await self.db.delete_collection(temp.name)

        async def update_batch() -> None:
            temp = await self.get_tmp_collection(change_id)
            log.info(f"Batch update: use temp collection {temp.name}")
            await store_to_tmp_collection(temp)

        if is_batch:
            await update_batch()
            await self.refresh_marked_update(change_id)
        else:
            await update_via_temp_collection()
        log.debug("Persist update done.")

    async def commit_batch_update(self, batch_id: str) -> None:
        temp_table = await self.get_tmp_collection(batch_id, False)
        await self.move_temp_to_proper(batch_id, temp_table.name)
        await self.db.delete_collection(temp_table.name)

    async def abort_update(self, batch_id: str) -> None:
        try:
            temp_table = await self.get_tmp_collection(batch_id, False)
            await self.db.delete_collection(temp_table.name)
        except NoSuchChangeError:
            pass
        await self.delete_marked_update(batch_id)

    async def to_query(self, query_model: QueryModel, with_edges: bool = False) -> Tuple[str, Json]:
        return arango_query.to_query(self, query_model, with_edges)

    async def insert_genesis_data(self) -> None:
        root_data = {"kind": "graph_root", "name": "root"}
        sha = GraphBuilder.content_hash(root_data)
        root_node = {"_key": "root", "id": "root", Section.reported: root_data, "kinds": ["graph_root"], "hash": sha}
        try:
            await self.db.insert(self.vertex_name, root_node)
        except Exception:
            # ignore if the root not is already created
            return None

    async def create_update_schema(self, init_with_data: bool = True, to_snapshot: bool = False) -> None:
        db = self.db

        async def create_update_graph(
            graph_name: GraphName, vertex_name: str, edge_name: str
        ) -> Tuple[Graph, VertexCollection, EdgeCollection]:
            graph = db.graph(graph_name) if await db.has_graph(graph_name) else await db.create_graph(graph_name)
            vertex_collection = (
                graph.vertex_collection(vertex_name)
                if await db.has_vertex_collection(graph_name, vertex_name)
                else await db.create_vertex_collection(graph_name, vertex_name)
            )
            edge_collection = (
                graph.edge_collection(edge_name)
                if await db.has_edge_definition(graph_name, edge_name)
                else await db.create_edge_definition(graph_name, edge_name, [vertex_name], [vertex_name])
            )
            return graph, vertex_collection, edge_collection

        def create_node_indexes(nodes: VertexCollection) -> None:
            node_idxes = {idx["name"]: idx for idx in cast(List[Json], nodes.indexes())}
            # this index will hold all the necessary data to query for an update (index only query)
            if "update_nodes_ref_id" not in node_idxes:
                log.info(f"Add index update_nodes_ref_id on {nodes.name}")
                nodes.add_persistent_index(
                    # if _key would be defined as first property, the optimizer would use it in case
                    # a simple id() query would be executed.
                    ["refs.cloud_id", "refs.account_id", "refs.region_id", "hash", "created", "_key"],
                    sparse=False,
                    name="update_nodes_ref_id",
                )

            if "kinds_id_name_ctime" not in node_idxes:
                nodes.add_persistent_index(
                    ["kinds[*]", "reported.id", "reported.name", "reported.ctime"],
                    sparse=False,
                    name="kinds_id_name_ctime",
                )

        def create_update_collection_indexes(progress: StandardCollection, node_history: StandardCollection) -> None:
            # progress indexes ------
            progress_idxes = {idx["name"]: idx for idx in cast(List[Json], progress.indexes())}
            if "parent_nodes" not in progress_idxes:
                log.info(f"Add index parent_nodes on {progress.name}")
                progress.add_persistent_index(["parent_nodes[*]"], name="parent_nodes")
            if "root_nodes" not in progress_idxes:
                log.info(f"Add index root_nodes on {progress.name}")
                progress.add_persistent_index(["root_nodes[*]"], name="root_nodes")
            # history indexes ------
            node_history_indexes = {idx["name"]: idx for idx in cast(List[Json], node_history.indexes())}
            if "history_access" not in node_history_indexes:
                node_history.add_persistent_index(
                    ["change", "changed_at", "kinds[*]", "reported.id", "reported.name", "reported.ctime"],
                    sparse=False,
                    name="history_access",
                )
            if "ttl_index" not in node_history_indexes:
                node_history.add_ttl_index(["changed"], int(timedelta(days=14).total_seconds()), name="ttl_index")

        def create_update_edge_indexes(edges: EdgeCollection) -> None:
            edge_idxes = {idx["name"]: idx for idx in cast(List[Json], edges.indexes())}
            # this index will hold all the necessary data to query for an update (index only query)
            if "update_edges_ref_id" not in edge_idxes:
                log.info(f"Add index update_edges_ref_id on {edges.name}")
                edges.add_persistent_index(
                    ["_key", "_from", "_to", "refs.cloud_id", "refs.account_id", "refs.region_id"],
                    sparse=False,
                    name="update_edges_ref_id",
                )
            outer_edge_ts_index_name = "outer_edge_timestamp_index"
            if outer_edge_ts_index_name not in edge_idxes:
                log.info(f"Add index {outer_edge_ts_index_name} on {edges.name}")
                edges.add_persistent_index(["outer_edge_ts"], sparse=True, name=outer_edge_ts_index_name)

        async def create_collection(name: str) -> StandardCollection:
            return db.collection(name) if await db.has_collection(name) else await db.create_collection(name)

        async def create_update_views(nodes: VertexCollection) -> None:
            name = f"search_{nodes.name}"
            prop = "flat"  # only the flat property is indexes

            # make sure the delimited analyzer exists
            try:
                db.db.analyzer("delimited")
            except AnalyzerGetError:
                db.db.create_analyzer(
                    "delimited",
                    "pipeline",
                    {
                        "pipeline": [
                            # lower case (leaving accents as is)
                            {"type": "norm", "properties": {"locale": "en_US.utf-8", "accent": True, "case": "lower"}},
                            # split the input
                            *[{"type": "delimiter", "properties": {"delimiter": a}} for a in fulltext_delimiter],
                            # remove empty strings
                            {"type": "stopwords", "properties": {"stopwords": [""], "hex": False}},
                        ]
                    },
                    ["frequency", "norm", "position"],
                )

            views = {view["name"]: view for view in await db.views()}
            if name not in views:
                log.info(f"Create view {name}")
                await db.create_view(
                    name,
                    "arangosearch",
                    {
                        "links": {nodes.name: {"fields": {prop: {"analyzers": ["delimited"]}}}},
                        "primarySort": [{"field": prop, "direction": "desc"}],
                        "inBackground": True,  # note: this setting only applies when the view is created
                    },
                )

        for edge_type in EdgeTypes.all:
            edge_type_name = self.edge_collection(edge_type)
            await create_update_graph(self.name, self.vertex_name, edge_type_name)

        vertex = db.graph(self.name).vertex_collection(self.vertex_name)

        if to_snapshot:
            # since the snapshots are immutable, we don't in_progress or node_history collections
            # we only create the indexes on the vertex collection
            create_node_indexes(vertex)
        else:
            in_progress = await create_collection(self.in_progress)
            node_history_collection = await create_collection(self.node_history)
            create_node_indexes(vertex)
            create_update_collection_indexes(in_progress, node_history_collection)
            await self.usage_db.create_update_schema()

        for edge_type in EdgeTypes.all:
            edge_collection = db.graph(self.name).edge_collection(self.edge_collection(edge_type))
            create_update_edge_indexes(edge_collection)

        await create_update_views(vertex)
        if init_with_data:
            await self.insert_genesis_data()

    async def copy_graph(self, to_graph: GraphName, to_snapshot: bool = False) -> GraphDB:
        if await self.db.has_graph(to_graph):
            raise ValueError(f"Graph {to_graph} already exists")

        new_graph_db = ArangoGraphDB(db=self.db, name=to_graph, adjust_node=self.node_adjuster, config=self.config)

        # collection creation can't be a part of a transaction so we do that first
        # we simply reuse the existing create_update_schema method but do not insert any genesis data
        async def create_new_collections(new_db: ArangoGraphDB, to_snapshot: bool) -> None:
            await new_db.create_update_schema(init_with_data=False, to_snapshot=to_snapshot)

        # we want to have a consistent snapshot view of the graph
        async def copy_data() -> None:
            old_vertex = self.vertex_name
            old_default_edge = self.edge_collection(EdgeTypes.default)
            old_delete_edge = self.edge_collection(EdgeTypes.delete)

            new_vertex = new_graph_db.vertex_name
            new_default_edge = new_graph_db.edge_collection(EdgeTypes.default)
            new_delete_edge = new_graph_db.edge_collection(EdgeTypes.delete)

            tx = f"""
                function () {{
                    const db = require('@arangodb').db;

                    db._query(
                        query="FOR vertex IN @@old_vertex "
                        + "LET clean_vertex = UNSET(vertex, '_id', '_rev') "
                        + "INSERT clean_vertex INTO @@new_vertex",
                        bindVars= {{
                            "@old_vertex": "{old_vertex}",
                            "@new_vertex": "{new_vertex}"
                        }}
                    );

                    db._query(
                        query="FOR edge IN @@old_default_edge "
                        + "LET clean_edge = UNSET(edge, '_id', '_rev') "
                        + "LET from = CONCAT(@new_vertex, '/', PARSE_IDENTIFIER(edge._from)['key']) "
                        + "LET to = CONCAT(@new_vertex, '/', PARSE_IDENTIFIER(edge._to)['key']) "
                        + "INSERT MERGE(clean_edge, {{_from: from, _to: to}}) INTO @@new_default_edge",
                        bindVars= {{
                            "@old_default_edge": "{old_default_edge}",
                            "@new_default_edge": "{new_default_edge}",
                            "new_vertex": "{new_vertex}"
                        }}
                    );

                    db._query(
                        query="FOR edge IN @@old_delete_edge "
                        + "LET clean_edge = UNSET(edge, '_id', '_rev') "
                        + "LET from = CONCAT(@new_vertex, '/', PARSE_IDENTIFIER(edge._from)['key']) "
                        + "LET to = CONCAT(@new_vertex, '/', PARSE_IDENTIFIER(edge._to)['key']) "
                        + "INSERT MERGE(clean_edge, {{_from: from, _to: to}}) INTO @@new_delete_edge",
                        bindVars= {{
                            "@old_delete_edge": "{old_delete_edge}",
                            "@new_delete_edge": "{new_delete_edge}",
                            "new_vertex": "{new_vertex}"
                        }}
                    );
                }}
            """
            await self.db.execute_transaction(
                tx,
                read=[old_vertex, old_default_edge, old_delete_edge],
                write=[new_vertex, new_default_edge, new_delete_edge],
            )

        await create_new_collections(new_graph_db, to_snapshot=to_snapshot)
        await copy_data()

        return cast(GraphDB, new_graph_db)

    async def insert_usage_data(self, data: List[UsageDatapoint]) -> None:
        if self.name.startswith("snapshot"):
            raise ValueError("Cannot insert usage data into a snapshot graph")
        await self.usage_db.update_many(data)

    @staticmethod
    def db_edge_key(from_node: str, to_node: str) -> str:
        return str(uuid.uuid5(uuid.NAMESPACE_DNS, f"{from_node}:{to_node}"))

    # parameter: rid
    # return: the complete document
    def query_node_by_id(self) -> str:
        return f"""
      FOR resource in `{self.vertex_name}`
      FILTER resource._key==@rid
      LIMIT 1
      RETURN resource
      """

    def query_update_nodes(self, merge_node_kind: str) -> str:
        return f"""
        FOR a IN `{self.vertex_name}`
        FILTER a.refs.{merge_node_kind}_id==@update_id
        RETURN {{_key: a._key, hash:a.hash, created:a.created}}
        """

    def query_update_edges(self, edge_type: EdgeType, merge_node_kind: str) -> str:
        collection = self.edge_collection(edge_type)
        return f"""
        FOR a IN `{collection}`
        FILTER a.refs.{merge_node_kind}_id==@update_id
        RETURN {{_key: a._key, _from: a._from, _to: a._to}}
        """

    def query_update_nodes_by_ids(self) -> str:
        return f"""
        FOR a IN `{self.vertex_name}`
        FILTER a._key IN @ids
        RETURN {{_key: a._key, hash:a.hash, created:a.created}}
        """

    def query_update_edges_by_ids(self, edge_type: EdgeType) -> str:
        collection = self.edge_collection(edge_type)
        return f"""
        FOR a IN `{collection}`
        FILTER a._key in @ids
        RETURN {{_key: a._key, _from: a._from, _to: a._to}}
        """

    def query_update_parent_linked(self) -> str:
        return f"""
        FOR a IN `{self.edge_collection(EdgeTypes.default)}`
        FILTER a._from==@from and a._to==@to
        RETURN true
        """

    def query_update_desired_metadata_many(self, section: str) -> str:
        return f"""
        FOR a IN `{self.vertex_name}`
        FILTER a._key in @node_ids
        UPDATE a with {{ "{section}": @patch }} IN `{self.vertex_name}`
        RETURN NEW
        """

    def query_delete_desired_metadata_many(self, section: str) -> str:
        return f"""
        FOR a IN `{self.vertex_name}`
        FILTER a._key in @node_ids
        REPLACE a with UNSET(a, "{section}") IN `{self.vertex_name}`
        RETURN NEW
        """

    def query_count_direct_children(self) -> str:
        return f"""
        FOR pn in `{self.vertex_name}` FILTER pn._key==@rid LIMIT 1
        FOR c IN 1..1 OUTBOUND pn {self.edge_collection(EdgeTypes.default)} COLLECT WITH COUNT INTO length
        RETURN length
        """

    def query_active_updates(self) -> str:
        return f"""
        FOR c IN `{self.in_progress}`
        RETURN {{id: c.change, created: c.created, affected_nodes: c.root_node_ids, is_batch: c.is_batch}}
        """  # noqa: E501

    def query_active_change(self) -> str:
        return f"""
        FOR change IN `{self.in_progress}`
        FILTER @root_node_ids any in change.parent_node_ids OR @root_node_ids any in change.root_node_ids
        RETURN change
        """

    def update_active_change(self) -> str:
        return f"""
        FOR d in `{self.in_progress}`
        FILTER d.change == @change
        UPDATE d WITH {{created: DATE_ISO8601(DATE_NOW())}} in `{self.in_progress}`
        """  # noqa: E501


class EventGraphDB(GraphDB):
    def __init__(self, real: ArangoGraphDB, event_sender: AnalyticsEventSender):
        self.real = real
        self.event_sender = event_sender
        self.graph_name = real.name

    @property
    def name(self) -> GraphName:
        return self.real.name

    async def get_node(self, model: Model, node_id: NodeId) -> Optional[Json]:
        return await self.real.get_node(model, node_id)

    async def create_node(self, model: Model, node_id: NodeId, data: Json, under_node_id: NodeId) -> Json:
        result = await self.real.create_node(model, node_id, data, under_node_id)
        await self.event_sender.core_event(CoreEvent.NodeCreated, {"graph": self.graph_name})
        return result

    async def update_deferred_edges(self, edges: List[Tuple[NodeId, NodeId, str]], ts: datetime) -> Tuple[int, int]:
        updated, deleted = await self.real.update_deferred_edges(edges, ts)
        await self.event_sender.core_event(CoreEvent.DeferredEdgesUpdated, updated=updated, deleted=deleted)
        return updated, deleted

    async def update_node(
        self, model: Model, node_id: NodeId, patch_or_replace: Json, replace: bool, section: Optional[str]
    ) -> Json:
        result = await self.real.update_node(model, node_id, patch_or_replace, replace, section)
        await self.event_sender.core_event(CoreEvent.NodeUpdated, {"graph": self.graph_name, "section": section})
        return result

    async def delete_node(self, node_id: NodeId) -> None:
        await self.real.delete_node(node_id)
        await self.event_sender.core_event(CoreEvent.NodeDeleted, {"graph": self.graph_name})

    def update_nodes(
        self, model: Model, patches_by_id: Dict[NodeId, Json], **kwargs: Any
    ) -> AsyncGenerator[Json, None]:
        return self.real.update_nodes(model, patches_by_id, **kwargs)

    async def update_nodes_desired(
        self, model: Model, patch: Json, node_ids: List[NodeId], **kwargs: Any
    ) -> AsyncGenerator[Json, None]:
        result = self.real.update_nodes_desired(model, patch, node_ids, **kwargs)
        await self.event_sender.core_event(
            CoreEvent.NodesDesiredUpdated, {"graph": self.graph_name}, updated=len(node_ids)
        )
        async for a in result:
            yield a

    async def update_nodes_metadata(
        self, model: Model, patch: Json, node_ids: List[NodeId], **kwargs: Any
    ) -> AsyncGenerator[Json, None]:
        result = self.real.update_nodes_metadata(model, patch, node_ids, **kwargs)
        await self.event_sender.core_event(
            CoreEvent.NodesMetadataUpdated, {"graph": self.graph_name}, updated=len(node_ids)
        )
        async for a in result:
            yield a

    async def merge_graph(
        self,
        graph_to_merge: MultiDiGraph,
        model: Model,
        maybe_change_id: Optional[str] = None,
        is_batch: bool = False,
    ) -> Tuple[List[str], GraphUpdate]:
        roots, info = await self.real.merge_graph(graph_to_merge, model, maybe_change_id, is_batch)
        root_counter: Dict[str, int] = {}
        for root in roots:
            root_node = graph_to_merge.nodes[root]
            rep_id = value_in_path_get(root_node, NodePath.reported_id, root)
            root_counter[f"node_count_{rep_id}.total"] = value_in_path_get(root_node, NodePath.descendant_count, 0)
            summary: Dict[str, int] = value_in_path_get(root_node, NodePath.descendant_summary, {})
            for nd_name, nd_count in summary.items():
                root_counter[f"node_count_{rep_id}.{nd_name}"] = nd_count

        # Filter the cloud nodes and get the name
        provider_names = [
            value_in_path_get(data, NodePath.reported_name, node_id)
            for node_id, data in graph_to_merge.nodes(data=True)
            if "cloud" in data.get("kinds", [])
        ]
        event_data: Dict[str, JsonElement] = {"graph": self.graph_name, "providers": provider_names, "batch": is_batch}

        kind = CoreEvent.BatchUpdateGraphMerged if is_batch else CoreEvent.GraphMerged
        await self.event_sender.core_event(
            kind,
            event_data,
            nodes=len(graph_to_merge.nodes),
            edges=len(graph_to_merge.edges),
            updated_roots=len(roots),
            updated=info.all_changes(),
            nodes_updated=info.nodes_updated,
            nodes_deleted=info.nodes_deleted,
            nodes_created=info.nodes_created,
            edges_created=info.edges_created,
            edges_updated=info.edges_updated,
            edges_deleted=info.edges_deleted,
            **root_counter,
        )
        return roots, info

    async def list_in_progress_updates(self) -> List[Json]:
        return await self.real.list_in_progress_updates()

    async def commit_batch_update(self, batch_id: str) -> None:
        info = first(lambda x: x["id"] == batch_id, await self.real.list_in_progress_updates())
        await self.real.commit_batch_update(batch_id)
        await self.event_sender.core_event(CoreEvent.BatchUpdateCommitted, {"graph": self.graph_name, "batch": info})

    async def abort_update(self, batch_id: str) -> None:
        info = first(lambda x: x["id"] == batch_id, await self.real.list_in_progress_updates())
        await self.real.abort_update(batch_id)
        await self.event_sender.core_event(CoreEvent.BatchUpdateAborted, {"graph": self.graph_name, "batch": info})

    async def search_list(
        self, query: QueryModel, with_count: bool = False, timeout: Optional[timedelta] = None, **kwargs: Any
    ) -> AsyncCursorContext:
        return await self.real.search_list(query, with_count, timeout, **kwargs)

    async def search_history(
        self,
        query: QueryModel,
        change: Optional[HistoryChange] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        with_count: bool = False,
        timeout: Optional[timedelta] = None,
        **kwargs: Any,
    ) -> AsyncCursorContext:
        counters, context = query.query.analytics()
        await self.event_sender.core_event(CoreEvent.HistoryQuery, context, **counters)
        return await self.real.search_history(query, change, before, after, with_count, timeout, **kwargs)

    async def search_graph_gen(
        self, query: QueryModel, with_count: bool = False, timeout: Optional[timedelta] = None
    ) -> AsyncCursorContext:
        return await self.real.search_graph_gen(query, with_count, timeout)

    async def search_aggregation(self, query: QueryModel) -> AsyncCursorContext:
        return await self.real.search_aggregation(query)

    async def search_graph(self, query: QueryModel) -> MultiDiGraph:
        return await self.real.search_graph(query)

    async def explain(self, query: QueryModel, with_edges: bool = False) -> EstimatedSearchCost:
        return await self.real.explain(query)

    async def wipe(self) -> None:
        await self.real.wipe()
        await self.event_sender.core_event(CoreEvent.GraphDBWiped, {"graph": self.graph_name})

    async def to_query(self, query_model: QueryModel, with_edges: bool = False) -> Tuple[str, Json]:
        return await self.real.to_query(query_model, with_edges)

    async def create_update_schema(self) -> None:
        await self.real.create_update_schema()

    async def copy_graph(self, to_graph: GraphName, to_snapshot: bool = False) -> GraphDB:
        await self.event_sender.core_event(
            CoreEvent.GraphCopied,
            {"graph": self.graph_name, "to_graph": to_graph},
        )
        return await self.real.copy_graph(to_graph, to_snapshot=to_snapshot)

    async def insert_usage_data(self, data: List[UsageDatapoint]) -> None:
        await self.real.insert_usage_data(data)



import asyncio
from typing import List

import pytest
from arango.database import StandardDatabase

from resotocore.analytics import AnalyticsEventSender, InMemoryEventSender
from resotocore.db import subscriberdb
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import EventEntityDb
from resotocore.db.subscriberdb import SubscriberDb, EventSubscriberDb
from resotocore.ids import SubscriberId
from resotocore.task.model import Subscriber, Subscription


@pytest.fixture
async def subscriber_db(test_db: StandardDatabase) -> SubscriberDb:
    async_db = AsyncArangoDB(test_db)
    subscriber_db = subscriberdb.subscriber_db(async_db, "subscriber")
    await subscriber_db.create_update_schema()
    await subscriber_db.wipe()
    return subscriber_db


@pytest.fixture
def event_db(subscriber_db: SubscriberDb, event_sender: AnalyticsEventSender) -> EventSubscriberDb:
    return EventEntityDb(subscriber_db, event_sender, "subscriber")


@pytest.fixture
def subscribers() -> List[Subscriber]:
    subs = [Subscription("foo", True) for _ in range(0, 10)]
    return [Subscriber.from_list(SubscriberId(str(a)), subs) for a in range(0, 10)]


@pytest.mark.asyncio
async def test_load(subscriber_db: SubscriberDb, subscribers: List[Subscriber]) -> None:
    await subscriber_db.update_many(subscribers)
    loaded = [sub async for sub in subscriber_db.all()]
    assert subscribers.sort() == loaded.sort()


@pytest.mark.asyncio
async def test_update(subscriber_db: SubscriberDb, subscribers: List[Subscriber]) -> None:
    # multiple updates should work as expected
    await subscriber_db.update_many(subscribers)
    await subscriber_db.update_many(subscribers)
    await subscriber_db.update_many(subscribers)
    loaded = [sub async for sub in subscriber_db.all()]
    assert subscribers.sort() == loaded.sort()


@pytest.mark.asyncio
async def test_delete(subscriber_db: SubscriberDb, subscribers: List[Subscriber]) -> None:
    await subscriber_db.update_many(subscribers)
    remaining = list(subscribers)
    for _ in subscribers:
        sub = remaining.pop()
        await subscriber_db.delete_value(sub)
        loaded = [sub async for sub in subscriber_db.all()]
        assert remaining.sort() == loaded.sort()
    assert len([sub async for sub in subscriber_db.all()]) == 0


@pytest.mark.asyncio
async def test_events(
    event_db: EventSubscriberDb, subscribers: List[Subscriber], event_sender: InMemoryEventSender
) -> None:
    # 2 times update
    await event_db.update_many(subscribers)
    await event_db.update_many(subscribers)
    # 6 times delete
    for sub in subscribers:
        await event_db.delete_value(sub)
    # make sure all events will arrive
    await asyncio.sleep(0.1)
    # ensure the correct count and order of events
    assert [a.kind for a in event_sender.events] == ["subscriber-updated-many"] * 2 + ["subscriber-deleted"] * 10



from asyncio import sleep
from contextlib import suppress, asynccontextmanager
from multiprocessing import Process
from typing import AsyncIterator, List, Optional
from pathlib import Path
import tempfile


import pytest
from _pytest.fixtures import fixture
from aiohttp import ClientSession, MultipartReader
from networkx import MultiDiGraph
from datetime import timedelta
from resotoclient import models as rc
from resotoclient.async_client import ResotoClient

from resotolib.utils import get_free_port
from tests.resotocore import create_graph
from resotocore.__main__ import run
from resotocore.analytics import AnalyticsEvent
from resotocore.db.db_access import DbAccess
from resotocore.model.model import predefined_kinds, Kind
from resotocore.model.typed_model import to_js
from resotocore.util import rnd_str, AccessJson, utc, utc_str
from resotocore.ids import GraphName
from resotoclient.json_utils import json_loadb
from resotoclient.models import JsObject


def graph_to_json(graph: MultiDiGraph) -> List[rc.JsObject]:
    ga: List[rc.JsValue] = [{**node, "type": "node"} for _, node in graph.nodes(data=True)]
    for from_node, to_node, data in graph.edges(data=True):
        ga.append({"type": "edge", "from": from_node, "to": to_node, "edge_type": data["edge_type"]})
    return ga


@fixture
async def core_client(
    client_session: ClientSession, foo_kinds: List[Kind], db_access: DbAccess
) -> AsyncIterator[ResotoClient]:
    async with create_core_client(client_session, foo_kinds, db_access, None) as client:
        yield client


@fixture
async def core_client_with_psk(
    client_session: ClientSession, foo_kinds: List[Kind], db_access: DbAccess
) -> AsyncIterator[ResotoClient]:
    async with create_core_client(client_session, foo_kinds, db_access, psk="test") as client:
        yield client


@asynccontextmanager
async def create_core_client(
    client_session: ClientSession,
    foo_kinds: List[Kind],
    db_access: DbAccess,
    psk: Optional[str] = None,
) -> AsyncIterator[ResotoClient]:
    """
    Note: adding this fixture to a test: a complete resotocore process is started.
          The fixture ensures that the underlying process has entered the ready state.
          It also ensures to clean up the process, when the test is done.
    """
    http_port = get_free_port()  # use a different port than the default one
    additional_args = ["--psk", psk] if psk else []

    # wipe and cleanly import the test model
    for graph_name in [g, "test", "hello", "bonjour", "foo", "resoto"]:
        db = await db_access.get_graph_model_db(GraphName(graph_name))
        await db.create_update_schema()
        await db.wipe()
        await db.update_many(foo_kinds)

    config_dir = tempfile.TemporaryDirectory()
    # todo: do not restart after the config override was loaded for the very first time and uncomment this part

    config_path = Path(config_dir.name) / "test_override_config_id.yaml"

    with config_path.open("w") as override_config:
        override_config.write(
            """
l1:
    l2: 42
        """
        )

    process = Process(
        target=run,
        args=(
            [
                "--graphdb-database",
                "test",
                "--graphdb-username",
                "test",
                "--graphdb-password",
                "test",
                "--debug",
                "--analytics-opt-out",
                "--override",
                f"resotocore.api.https_port=null",
                f"resotocore.api.http_port={http_port}",
                "resotocore.api.web_hosts=0.0.0.0",
                "--override-path",
                str(config_path),
                *additional_args,
            ],
        ),
    )
    process.start()
    ready = False
    count = 20
    while not ready:
        await sleep(0.5)
        with suppress(Exception):
            async with client_session.get(f"http://localhost:{http_port}/system/ready"):
                ready = True
        count -= 1
        if count == 0:
            raise AssertionError("Process does not came up as expected")
    async with ResotoClient(f"http://localhost:{http_port}", psk=psk) as client:
        yield client
    # terminate the process
    process.terminate()
    process.join(5)
    # if it is still running, kill it
    if process.is_alive():
        process.kill()
        process.join()
    process.close()
    config_dir.cleanup()


g = "graphtest"


@pytest.mark.asyncio
async def test_system_api(core_client: ResotoClient, client_session: ClientSession) -> None:
    assert await core_client.ping() == "pong"
    assert await core_client.ready() == "ok"
    # make sure we get redirected to the api docs
    async with client_session.get(core_client.resotocore_url, allow_redirects=False) as r:
        assert r.headers["location"] == "/ui/index.html"
    # analytics events can be sent to the server
    events = [AnalyticsEvent("test", "test.event", {"foo": "bar"}, {"counter": 1}, utc())]
    async with client_session.post(core_client.resotocore_url + "/analytics", json=to_js(events)) as r:
        assert r.status == 204


@pytest.mark.asyncio
async def test_model_api(core_client: ResotoClient, client_session: ClientSession) -> None:
    # GET /model
    assert len((await core_client.model()).kinds) >= len(predefined_kinds)

    # PATCH /model
    string_kind: rc.Kind = rc.Kind(fqn="only_three", runtime_kind="string", properties=None, bases=None)
    setattr(string_kind, "min_length", 3)
    setattr(string_kind, "max_length", 3)

    prop = rc.Property(name="ot", kind="only_three", required=False, metadata={"len": 3})
    complex_kind: rc.Kind = rc.Kind(fqn="test_cpl", runtime_kind=None, properties=[prop], bases=None, metadata={"a": 1})
    setattr(complex_kind, "allow_unknown_props", False)

    update = await core_client.update_model([string_kind, complex_kind])
    assert update.kinds["only_three"].runtime_kind == "string"
    assert update.kinds["test_cpl"].metadata["a"] == 1
    assert update.kinds["test_cpl"].properties[0].metadata["len"] == 3

    # GET /model/uml
    async with client_session.get(core_client.resotocore_url + "/model/uml", params={"output": "puml"}) as r:
        assert r.status == 200
        assert r.headers["content-type"] == "text/plain"
        puml = await r.text()
        assert puml.startswith("@startuml")
        assert puml.endswith("@enduml")


@pytest.mark.asyncio
async def test_graph_api(core_client: ResotoClient) -> None:
    # make sure we have a clean slate
    with suppress(Exception):
        await core_client.delete_graph(g)

    # create a new graph
    graph = AccessJson(await core_client.create_graph(g))
    assert graph.id == "root"
    assert graph.reported.kind == "graph_root"

    # list all graphs
    graphs = await core_client.list_graphs()
    assert g in graphs

    # get one specific graph
    graph: AccessJson = AccessJson(await core_client.get_graph(g))  # type: ignore
    assert graph.id == "root"
    assert graph.reported.kind == "graph_root"

    # wipe the data in the graph
    assert await core_client.delete_graph(g, truncate=True) == "Graph truncated."
    assert g in await core_client.list_graphs()

    # create a node in the graph
    uid = rnd_str()
    node = AccessJson(
        await core_client.create_node("root", uid, {"identifier": uid, "kind": "child", "name": "max"}, g)
    )
    assert node.id == uid
    assert node.reported.name == "max"

    # update a node in the graph
    node = AccessJson(await core_client.patch_node(uid, {"name": "moritz"}, "reported", g))
    assert node.id == uid
    assert node.reported.name == "moritz"

    # get the node
    node = AccessJson(await core_client.get_node(uid, g))
    assert node.id == uid
    assert node.reported.name == "moritz"

    # delete the node
    await core_client.delete_node(uid, g)
    with pytest.raises(AttributeError):
        # node can not be found
        await core_client.get_node(uid, g)

    # merge a complete graph
    merged = await core_client.merge_graph(graph_to_json(create_graph("test")), g)
    assert merged == rc.GraphUpdate(112, 1, 0, 212, 0, 0)

    # batch graph update and commit
    batch1_id, batch1_info = await core_client.add_to_batch(graph_to_json(create_graph("hello")), "batch1", g)
    assert batch1_info == rc.GraphUpdate(0, 100, 0, 0, 0, 0)
    assert batch1_id == "batch1"
    batch_infos = AccessJson.wrap_list(await core_client.list_batches(g))
    assert len(batch_infos) == 1
    # assert batch_infos[0].id == batch1_id
    assert batch_infos[0].affected_nodes == ["collector"]  # replace node
    assert batch_infos[0].is_batch is True
    await core_client.commit_batch(batch1_id, g)

    # batch graph update and abort
    batch2_id, batch2_info = await core_client.add_to_batch(graph_to_json(create_graph("bonjour")), "batch2", g)
    assert batch2_info == rc.GraphUpdate(0, 100, 0, 0, 0, 0)
    assert batch2_id == "batch2"
    await core_client.abort_batch(batch2_id, g)

    # update nodes
    update = [{"id": node["id"], "reported": {"name": "bruce"}} for _, node in create_graph("foo").nodes(data=True)]
    updated_nodes = await core_client.patch_nodes(update, g)
    assert len(updated_nodes) == 113
    for n in updated_nodes:
        assert n.get("reported", {}).get("name") == "bruce"

    # create the raw search
    raw = await core_client.search_graph_raw('id("3")', g)
    assert raw == {
        "query": "LET filter0 = (FOR m0 in `graphtest` FILTER m0._key == @b0  RETURN m0) "
        'FOR result in filter0 RETURN UNSET(result, ["flat"])',
        "bind_vars": {"b0": "3"},
    }

    # estimate the search
    cost = await core_client.search_graph_explain('id("3")', g)
    assert cost.full_collection_scan is False
    assert cost.rating == rc.EstimatedQueryCostRating.simple

    # search list
    result_list = [res async for res in core_client.search_list('id("3") -[0:]->', graph=g)]
    assert len(result_list) == 11  # one parent node and 10 child nodes
    assert result_list[0].get("id") == "3"  # first node is the parent node

    # search graph
    result_graph = [res async for res in core_client.search_graph('id("3") -[0:]->', graph=g)]
    assert len(result_graph) == 21  # 11 nodes + 10 edges
    assert result_list[0].get("id") == "3"  # first node is the parent node

    # search graph at specific timestamp
    async def search_graph_at(
        search: str, section: Optional[str] = "reported", graph: str = "resoto", at: Optional[str] = None
    ) -> AsyncIterator[JsObject]:
        params = {}
        if section:
            params["section"] = section
        if at:
            params["at"] = at
        response = await core_client._post(f"/graph/{graph}/search/graph", params=params, data=search, stream=True)
        if response.status_code == 200:
            async for line in response.async_iter_lines():
                yield json_loadb(line)
        else:
            raise AttributeError(await response.text())

    with pytest.raises(AttributeError):
        # no snapshots 420 weeks ago
        result = [
            res
            async for res in search_graph_at('id("3") -[0:]->', graph=g, at=(utc() - timedelta(weeks=420)).isoformat())
        ]
        assert len(result) == 0

    # create a snapshot
    async for _ in core_client.cli_execute("graph snapshot graphtest test_label"):
        pass
    # now we should see some snapshots
    result_graph = [res async for res in search_graph_at('id("3") -[0:]->', graph=g, at=utc_str())]
    assert len(result_graph) == 21  # 11 nodes + 10 edges

    # aggregate
    result_aggregate = core_client.search_aggregate("aggregate(kind as kind: sum(1) as count): all", graph=g)
    assert {r["group"]["kind"]: r["count"] async for r in result_aggregate} == {
        "bla": 100,
        "cloud": 1,
        "foo": 11,
        "graph_root": 1,
    }

    # delete the graph
    assert await core_client.delete_graph(g) == "Graph deleted."
    assert g not in await core_client.list_graphs()


@pytest.mark.asyncio
async def test_subscribers(core_client: ResotoClient) -> None:
    # provide a clean slate
    for subscriber in await core_client.subscribers():
        await core_client.delete_subscriber(subscriber.id)

    sub_id = rnd_str()

    # add subscription
    subscriber = await core_client.add_subscription(sub_id, rc.Subscription("test"))
    assert subscriber.id == sub_id
    assert len(subscriber.subscriptions) == 1
    assert subscriber.subscriptions["test"] is not None

    # delete subscription
    subscriber = await core_client.delete_subscription(sub_id, rc.Subscription("test"))
    assert subscriber.id == sub_id
    assert len(subscriber.subscriptions) == 0

    # update subscriber
    updated = await core_client.update_subscriber(sub_id, [rc.Subscription("test"), rc.Subscription("rest")])
    assert updated is not None
    assert updated.id == sub_id
    assert len(updated.subscriptions) == 2

    # subscriber for message type
    assert await core_client.subscribers_for_event("test") == [updated]
    assert await core_client.subscribers_for_event("rest") == [updated]
    assert await core_client.subscribers_for_event("does_not_exist") == []

    # get subscriber
    sub = await core_client.subscriber(sub_id)
    assert sub is not None


@pytest.mark.asyncio
async def test_cli(core_client: ResotoClient) -> None:
    # make sure we have a clean slate
    with suppress(Exception):
        await core_client.delete_graph(g)
    await core_client.create_graph(g)
    graph_update = graph_to_json(create_graph("test"))
    await core_client.merge_graph(graph_update, g)

    # evaluate search with count
    result = await core_client.cli_evaluate("search all | count kind", g)
    assert len(result) == 1
    parsed, to_execute = result[0]
    assert len(parsed.commands) == 2
    assert (parsed.commands[0].cmd, parsed.commands[1].cmd) == ("search", "count")
    assert len(to_execute) == 3
    assert (to_execute[0].get("cmd"), to_execute[1].get("cmd")) == ("execute_search", "aggregate_to_count")

    # execute search with count
    executed = [result async for result in core_client.cli_execute("search is(foo) or is(bla) | count kind", g)]
    assert executed == ["cloud: 1", "foo: 11", "bla: 100", "total matched: 112", "total unmatched: 0"]

    # execute multiple commands
    response = await core_client.cli_execute_raw("echo foo; echo bar; echo bla")
    reader: MultipartReader = MultipartReader.from_response(response.undrelying)  # type: ignore
    assert [await p.text() async for p in reader] == ['"foo"', '"bar"', '"bla"']

    # list all cli commands
    info = AccessJson(await core_client.cli_info())
    assert len(info.commands) == 44


@pytest.mark.asyncio
async def test_config(core_client: ResotoClient, foo_kinds: List[rc.Kind]) -> None:
    # make sure we have a clean slate
    async for config in core_client.configs():
        await core_client.delete_config(config)

    # define a config model
    model = await core_client.update_configs_model(foo_kinds)
    assert "foo" in model.kinds
    assert "bla" in model.kinds
    # get the config model again
    get_model = await core_client.get_configs_model()
    assert len(model.kinds) == len(get_model.kinds)

    # define config validation
    validation = rc.ConfigValidation("external.validated.config", external_validation=True)
    assert await core_client.put_config_validation(validation) == validation

    # get the config validation
    assert await core_client.get_config_validation(validation.id) == validation

    # put config
    cfg_id = rnd_str()

    # put a config with schema that is violated
    with pytest.raises(AttributeError) as ex:
        await core_client.put_config(cfg_id, {"foo": {"some_int": "abc"}})
    assert "Expected type int32 but got str" in str(ex.value)

    # put a config with schema that is violated, but turn validation off
    await core_client.put_config(cfg_id, {"foo": {"some_int": "abc"}}, validate=False)

    # set a simple state
    assert await core_client.put_config(cfg_id, {"a": 1}) == {"a": 1}

    # patch config
    assert await core_client.patch_config(cfg_id, {"a": 1}) == {"a": 1}
    assert await core_client.patch_config(cfg_id, {"b": 2}) == {"a": 1, "b": 2}
    assert await core_client.patch_config(cfg_id, {"c": 3}) == {"a": 1, "b": 2, "c": 3}

    # get config
    assert await core_client.config(cfg_id) == {"a": 1, "b": 2, "c": 3}

    # list configs
    assert [conf async for conf in core_client.configs()] == [cfg_id]

    # delete config
    await core_client.delete_config(cfg_id)
    assert [conf async for conf in core_client.configs()] == []

    cfg_override_id = "test_override_config_id"

    # set a simple state, the override should not be applied since
    # we want to get a DB value only
    put_result = await core_client.put_config(cfg_override_id, {"l1": {"l2": 1}})
    assert put_result == {"l1": {"l2": 1}}

    # get config, override should be applied
    with_overrides = await core_client.config(cfg_override_id)
    assert with_overrides == {"l1": {"l2": 42}}

    # get config with overrides in different section
    resp = await core_client._get(
        f"/config/{cfg_override_id}", params={"separate_overrides": "true", "include_raw_config": "true"}
    )
    json = await resp.json()
    assert json == {
        "config": {"l1": {"l2": 1}},
        "overrides": {"l1": {"l2": 42}},
        "raw_config": {"l1": {"l2": 1}},
    }

    # raw config is not sent by default
    resp = await core_client._get(f"/config/{cfg_override_id}", params={"separate_overrides": "true"})
    json = await resp.json()
    assert json == {
        "config": {"l1": {"l2": 1}},
        "overrides": {"l1": {"l2": 42}},
    }


@pytest.mark.asyncio
async def test_authorization(core_client_with_psk: ResotoClient, client_session: ClientSession) -> None:
    url = core_client_with_psk.resotocore_url
    # make sure all users are deleted
    await core_client_with_psk.delete_config("resoto.users")

    # Step 1: create first user ================================

    # getting a restricted resource returns a 401
    async with client_session.get(f"{url}/authorization/user", allow_redirects=False) as resp:
        assert resp.status == 401
    # go to the login page, which is the first user creation page
    async with client_session.get(f"{url}/login") as resp:
        assert resp.status == 200
        assert resp.content_type == "text/html"
        assert "/create-first-user" in await resp.text()
    # the form is submitted
    async with client_session.post(
        f"{url}/create-first-user",
        allow_redirects=False,
        data={
            "company": "some company",
            "fullname": "John Doe",
            "email": "test@test.de",
            "password": "test",
            "password_repeat": "test",
            "redirect": "/authorization/user",
        },
    ) as resp:
        assert resp.status == 303
        # a code is added to the redirect as request parameter
        user_with_code = resp.headers["Location"]
        assert "/authorization/user?code=" in user_with_code
    # the code is used to authenticate
    async with client_session.get(f"{url}{user_with_code}", allow_redirects=False) as resp:
        assert resp.status == 200
        auth_header = resp.headers["Authorization"]
        assert auth_header.startswith("Bearer ")
    # the auth header can be used for subsequent requests
    async with client_session.get(f"{url}/authorization/user", headers={"Authorization": auth_header}) as resp:
        assert resp.status == 200
        user = await resp.json()
        assert user["email"] == "test@test.de"
        assert user["roles"] == "admin"

    # Step 2: login with existing user ================================

    # subsequent interactions need to log in
    async with client_session.get(f"{url}/authorization/user") as resp:
        assert resp.status == 401
    async with client_session.get(f"{url}/login") as resp:
        assert resp.status == 200
        assert resp.content_type == "text/html"
        assert "/authenticate" in await resp.text()
    # the login form is submitted
    async with client_session.post(
        f"{url}/authenticate",
        allow_redirects=False,
        data={"email": "test@test.de", "password": "test", "redirect": "/authorization/user"},
    ) as resp:
        assert resp.status == 303
        # a code is added to the redirect as request parameter
        user_with_code = resp.headers["Location"]
        assert "/authorization/user?code=" in user_with_code
    # the code is used to authenticate
    async with client_session.get(f"{url}{user_with_code}", allow_redirects=False) as resp:
        assert resp.status == 200
        auth_header = resp.headers["Authorization"]
        assert auth_header.startswith("Bearer ")
    # the auth header can be used for subsequent requests
    async with client_session.get(f"{url}/authorization/user", headers={"Authorization": auth_header}) as resp:
        assert resp.status == 200

    # Step 3: renew the authorization header ================================
    # the auth header can be used for subsequent requests
    async with client_session.get(f"{url}/authorization/renew", headers={"Authorization": auth_header}) as resp:
        assert resp.status == 200
        assert resp.headers["Authorization"].startswith("Bearer ")



from __future__ import annotations

import asyncio
import logging
from asyncio import Task
from contextlib import suppress
from itertools import takewhile
from operator import attrgetter
from textwrap import dedent
from typing import Dict, List, Tuple, Union, Sequence
from typing import Optional, Any

from aiostream import stream
from aiostream.core import Stream
from attrs import evolve
from parsy import Parser
from rich.padding import Padding

from resotocore import version
from resotocore.analytics import CoreEvent
from resotocore.cli import cmd_with_args_parser, key_values_parser, T, Sink, args_values_parser
from resotocore.cli.command import (
    SearchPart,
    PredecessorsPart,
    SuccessorsPart,
    AncestorsPart,
    DescendantsPart,
    AggregateCommand,
    CountCommand,
    HeadCommand,
    TailCommand,
    SearchCLIPart,
    ExecuteSearchCommand,
    JobsCommand,
    WelcomeCommand,
    SortPart,
    LimitPart,
    HistoryPart,
    ReportCommand,
    WriteCommand,
)
from resotocore.dependencies import Dependencies
from resotocore.cli.model import (
    ParsedCommand,
    ParsedCommands,
    ExecutableCommand,
    ParsedCommandLine,
    CLICommand,
    InternalPart,
    CLIContext,
    CLI,
    EmptyContext,
    CLISource,
    NoTerminalOutput,
    OutputTransformer,
    PreserveOutputFormat,
    AliasTemplate,
    InfraAppAlias,
    ArgsInfo,
    ArgInfo,
    AliasTemplateParameter,
)
from resotocore.console_renderer import ConsoleRenderer
from resotocore.error import CLIParseError
from resotocore.model.typed_model import class_fqn
from resotocore.query.model import (
    Query,
    Navigation,
    AllTerm,
    Aggregate,
    AggregateVariable,
    AggregateVariableName,
    AggregateFunction,
    PathRoot,
    Limit,
    Sort,
)
from resotocore.query.query_parser import aggregate_parameter_parser, sort_args_p, limit_parser_direct
from resotocore.service import Service
from resotocore.types import JsonElement
from resotocore.util import group_by
from resotolib.parse_util import make_parser, pipe_p, semicolon_p

log = logging.getLogger(__name__)


@make_parser
def single_command_parser() -> Parser:
    parsed = yield cmd_with_args_parser
    cmd_args = [a.strip() for a in parsed.strip().split(" ", 1)]
    cmd, args = cmd_args if len(cmd_args) == 2 else (cmd_args[0], None)
    return ParsedCommand(cmd, args)


single_commands = single_command_parser.sep_by(pipe_p, min=1)


@make_parser
def command_line_parser() -> Parser:
    maybe_env = yield key_values_parser.optional()
    commands = yield single_commands
    return ParsedCommands(commands, maybe_env if maybe_env else {})


# multiple piped commands are separated by semicolon
multi_command_parser = command_line_parser.sep_by(semicolon_p)


class HelpCommand(CLICommand):
    """
    Usage: help [command]

    Parameter:
        command [optional]: if given shows the help for a specific command

    Show help text for a command or general help information.
    """

    def __init__(
        self,
        dependencies: Dependencies,
        parts: List[CLICommand],
        alias_names: Dict[str, str],
        alias_templates: Dict[str, AliasTemplate],
        infra_app_aliases: Dict[str, InfraAppAlias],
    ):
        super().__init__(dependencies, "misc", True)
        self.all_parts = {p.name: p for p in parts + [self]}
        self.parts = {p.name: p for p in parts + [self] if not isinstance(p, InternalPart)}
        self.alias_names = {a: n for a, n in alias_names.items() if n in self.parts and a not in self.parts}
        self.reverse_alias_names: Dict[str, List[str]] = {
            k: [e[0] for e in v] for k, v in group_by(lambda a: a[1], self.alias_names.items()).items()
        }
        self.alias_templates = alias_templates
        self.infra_app_aliases = infra_app_aliases

    @property
    def name(self) -> str:
        return "help"

    def info(self) -> str:
        return "Shows available commands, as well as help for any specific command."

    def args_info(self) -> ArgsInfo:
        return [ArgInfo(None, expects_value=True, value_hint="command")]

    def parse(self, arg: Optional[str] = None, ctx: CLIContext = EmptyContext, **kwargs: Any) -> CLISource:
        def placeholders() -> str:
            replacements = "\n".join(f"- `@{key}@` -> {value}" for key, value in CLI.replacements(**ctx.env).items())
            return ctx.render_console(f"## Valid placeholder string: \n\n{replacements}")

        def overview() -> str:
            all_parts = sorted(self.parts.values(), key=lambda p: p.name)
            parts = [p for p in all_parts if isinstance(p, CLICommand)]
            templates = list(sorted(self.alias_templates.values(), key=attrgetter("name")))
            alias_templates = "\n".join(f"- `{a.name}` - {a.info}" for a in templates)

            sorted_infra_app_aliases = list(sorted(self.infra_app_aliases.values(), key=attrgetter("name")))
            infra_app_aliases = "\n".join(f"- `{a.name}` - {a.description}" for a in sorted_infra_app_aliases)

            result = f"## Custom Commands \n{alias_templates}\n ## Infrastucture Apps \n{infra_app_aliases}\n"
            for category in ["search", "format", "action", "setup", "misc"]:
                result += f"\n\n## {category.capitalize()} Commands\n"
                for part in parts:
                    if part.category == category:
                        result += f"- `{part.name}` - {part.info()}\n"

            result += dedent(
                """

                 *Note* that you can pipe commands using the pipe character (|)
                 and chain multiple commands using the semicolon (;).

                 Use `help <command>` to show help for a specific command. \\
                 Use `help placeholders` to see the list of available placeholders.
                 """
            )
            headline = ctx.render_console(f"# resotocore CLI ({version()})")
            # ck mascot is centered (rendered if color is enabled)
            middle = (
                int((ctx.console_renderer.width - 22) / 2)
                if ctx.console_renderer is not None and ctx.console_renderer.width is not None
                else 0
            )
            logo = ctx.render_console(Padding(WelcomeCommand.ck, pad=(0, 0, 0, middle))) if ctx.supports_color() else ""
            return headline + logo + ctx.render_console(result)

        def help_command() -> Stream:
            if not arg:
                result = overview()
            elif arg == "placeholders":
                result = placeholders()
            elif arg in self.all_parts:
                maybe_aliases = self.reverse_alias_names.get(arg)
                result = ""
                if maybe_aliases:
                    result += f'{arg} can also invoked via: {", ".join(maybe_aliases)}\n\n'
                result += self.all_parts[arg].rendered_help(ctx)
            elif arg in self.alias_names:
                alias = self.alias_names[arg]
                explain = f"{arg} is an alias for {alias}\n\n"
                result = explain + self.all_parts[alias].rendered_help(ctx)
            elif arg in self.alias_templates:
                result = self.alias_templates[arg].rendered_help(ctx)
            elif arg in self.infra_app_aliases:
                result = self.infra_app_aliases[arg].rendered_help(ctx)
            else:
                result = f"No command found with this name: {arg}"

            return stream.just(result)

        return CLISource.single(help_command)


CLIArg = Tuple[CLICommand, Optional[str]]
# If no sort is defined in the part, we use this default sort order
DefaultSort = [Sort("/reported.kind"), Sort("/reported.name"), Sort("/reported.id")]
# Default sort order for history searches
HistorySort = [Sort("/changed_at"), Sort("/reported.kind"), Sort("/reported.name"), Sort("/reported.id")]


class CLIService(CLI, Service):
    """
    The CLI has a defined set of dependencies and knows a list if commands.
    A string can be parsed into a command line that can be executed based on the list of available commands.
    """

    def __init__(
        self, dependencies: Dependencies, parts: List[CLICommand], env: Dict[str, Any], alias_names: Dict[str, str]
    ):
        dependencies.extend(cli=self)
        alias_templates_list = [AliasTemplate.from_config(cmd) for cmd in dependencies.config.custom_commands.commands]
        alias_templates = {a.name: a for a in alias_templates_list}
        infra_app_aliases: Dict[str, InfraAppAlias] = {}
        help_cmd = HelpCommand(
            dependencies,
            parts,
            alias_names,
            alias_templates,
            infra_app_aliases,
        )
        cmds = {p.name: p for p in parts + [help_cmd]}
        alias_cmds = {alias: cmds[name] for alias, name in alias_names.items() if name in cmds and alias not in cmds}
        self.cli_env = env
        self.alias_names = alias_names
        self.__direct_commands = cmds
        self.__alias_commands = alias_cmds
        self.__commands: Dict[str, CLICommand] = {**cmds, **alias_cmds}
        self.__dependencies = dependencies
        self.__alias_templates = alias_templates
        self.__infra_app_aliases = infra_app_aliases
        self.reaper: Optional[Task[None]] = None

    @property
    def direct_commands(self) -> Dict[str, CLICommand]:
        return self.__direct_commands

    @property
    def alias_commands(self) -> Dict[str, CLICommand]:
        return self.__alias_commands

    @property
    def commands(self) -> Dict[str, CLICommand]:
        return self.__commands

    @property
    def env(self) -> Dict[str, Any]:
        return self.cli_env

    @property
    def dependencies(self) -> Dependencies:
        return self.__dependencies

    @property
    def alias_templates(self) -> Dict[str, AliasTemplate]:
        return self.__alias_templates

    @property
    def infra_app_aliases(self) -> Dict[str, InfraAppAlias]:
        return self.__infra_app_aliases

    def _no_name_conflict(self, name: str) -> bool:
        return (
            name not in self.direct_commands
            and name not in self.alias_commands
            and name not in self.alias_templates
            and name not in self.infra_app_aliases
        )

    def register_infra_app_alias(self, alias: InfraAppAlias) -> None:
        """
        Called when an infra app is registered.
        """
        if self._no_name_conflict(alias.name):
            self.__infra_app_aliases[alias.name] = alias

    def unregister_infra_app_alias(self, name: str) -> None:
        del self.__infra_app_aliases[name]

    def register_alias_template(self, template: AliasTemplate) -> None:
        """
        Called when something introduces a custom command.
        The registered templated will always override any existing template.
        """
        if self._no_name_conflict(template.name):
            self.alias_templates[template.name] = template

    def unregister_alias_template(self, name: str) -> None:
        """
        Called when something removes a custom command.
        """
        if name in self.alias_templates:
            del self.alias_templates[name]

    async def start(self) -> None:
        self.reaper = asyncio.create_task(self.reap_tasks())

    async def stop(self) -> None:
        if self.reaper:
            self.reaper.cancel()
            await asyncio.gather(self.reaper, return_exceptions=True)

        while not self.dependencies.forked_tasks.empty():
            task, _ = self.dependencies.forked_tasks.get_nowait()
            loop = asyncio.get_event_loop()
            with suppress(asyncio.CancelledError):
                if not task.done() or not task.cancelled():
                    task.cancel()
                loop.run_until_complete(task)

        await self.dependencies.stop()

    async def reap_tasks(self) -> None:
        while True:
            try:
                task, info = await self.dependencies.forked_tasks.get()
                try:
                    res = await task
                    log.info(f"Spawned task {info} completed with: {res}")
                except Exception as ex:
                    log.info(f"Spawned task {info} failed. Reason {ex}")
            except Exception as ex:
                log.warning(f"Error in main loop: {ex}")

    def command(
        self, name: str, arg: Optional[str] = None, ctx: CLIContext = EmptyContext, **kwargs: Any
    ) -> ExecutableCommand:
        """
        Create an executable command for given command name, args and context.
        :param name: the name of the command to execute (must be a known command)
        :param arg: the arg of the command (must be parsable by the command)
        :param ctx: the context of this command.
        :return: the ready to run executable command.
        :raises:
            CLIParseError: if the name of the command is not known, or the argument fails to parse.
        """
        if name in self.commands:
            command = self.commands[name]
            try:
                action = command.parse(arg, ctx, cmd_name=name, **kwargs)
                return ExecutableCommand(name, command, arg, action)
            except Exception as ex:
                raise CLIParseError(f"{name} can not parse arg {arg}. Reason: {ex}") from ex
        else:
            raise CLIParseError(f"Command >{name}< is not known. Typo?")

    async def create_query(
        self, commands: List[ExecutableCommand], ctx: CLIContext
    ) -> Tuple[Query, Dict[str, Any], List[ExecutableCommand]]:
        """
        Takes a list of query part commands and combine them to a single executable query command.
        This process can also introduce new commands that should run after the query is finished.
        Therefore, a list of executable commands is returned.
        :param commands: the incoming executable commands, which actions are all instances of SearchCLIPart.
        :param ctx: the context to execute within.
        :return: the resulting list of commands to execute.
        """

        # Pass parsed options to execute query
        # Multiple query commands are possible - so the dict is combined with every parsed query.
        parsed_options: Dict[str, Any] = {}

        async def parse_query(query_arg: str) -> Query:
            nonlocal parsed_options
            parsed, query_part = ExecuteSearchCommand.parse_known(query_arg)
            parsed_options = {**parsed_options, **parsed}
            # empty string is interpreted as no filter
            query_part = "all" if query_part.strip() == "" else query_part
            # section expansion is disabled here: it will happen on the final query after all parts have been combined
            return await self.dependencies.template_expander.parse_query(
                "".join(query_part), None, omit_section_expansion=True, **ctx.env
            )

        query: Query = Query.by(AllTerm())
        additional_commands: List[ExecutableCommand] = []
        # We need to remember the first head/tail, since tail will reverse the sort order
        first_head_tail_in_a_row: Optional[CLICommand] = None
        head_tail_keep_order = True
        for command in commands:
            part = command.command
            arg = command.arg if command.arg else ""
            if isinstance(part, SearchPart):
                query = query.combine(await parse_query(arg))
            elif isinstance(part, HistoryPart):
                parsed_options["history"] = True
                query = query.combine(await parse_query(arg))
                if not query.current_part.sort and not query.aggregate:
                    query = query.set_sort(*HistorySort)
            elif isinstance(part, SortPart):
                if query.current_part.sort == DefaultSort:
                    query = query.set_sort(*sort_args_p.parse(arg))
                else:
                    query = query.add_sort(*sort_args_p.parse(arg))
            elif isinstance(part, LimitPart):
                query = query.with_limit(limit_parser_direct.parse(arg))
            elif isinstance(part, PredecessorsPart):
                origin, edge = PredecessorsPart.parse_args(arg, ctx)
                query = query.traverse_in(origin, 1, edge)
            elif isinstance(part, SuccessorsPart):
                origin, edge = PredecessorsPart.parse_args(arg, ctx)
                query = query.traverse_out(origin, 1, edge)
            elif isinstance(part, AncestorsPart):
                origin, edge = PredecessorsPart.parse_args(arg, ctx)
                query = query.traverse_in(origin, Navigation.Max, edge)
            elif isinstance(part, DescendantsPart):
                origin, edge = PredecessorsPart.parse_args(arg, ctx)
                query = query.traverse_out(origin, Navigation.Max, edge)
            elif isinstance(part, AggregateCommand):
                group_vars, group_function_vars = aggregate_parameter_parser.parse(arg)
                query = evolve(query, aggregate=Aggregate(group_vars, group_function_vars))
            elif isinstance(part, CountCommand):
                # count command followed by a query: make it an aggregation
                # since the output of aggregation is not exactly the same as count
                # we also add the aggregate_to_count command after the query
                assert query.aggregate is None, "Can not combine aggregate and count!"
                group_by_var = [AggregateVariable(AggregateVariableName(arg), "name")] if arg else []
                aggregate = Aggregate(group_by_var, [AggregateFunction("sum", 1, [], "count")])
                # If the query should be explained, we want the output as is
                if "explain" not in parsed_options:
                    additional_commands.append(self.command("aggregate_to_count", None, ctx))
                query = evolve(query, aggregate=aggregate)
                query = query.set_sort(Sort(f"{PathRoot}count"))
            elif isinstance(part, HeadCommand):
                size = HeadCommand.parse_size(arg)
                limit = query.parts[0].limit or Limit(0, size)
                if first_head_tail_in_a_row and head_tail_keep_order:
                    query = query.with_limit(Limit(limit.offset, min(limit.length, size)))
                elif first_head_tail_in_a_row and not head_tail_keep_order:
                    length = min(limit.length, size)
                    query = query.with_limit(Limit(limit.offset + limit.length - length, length))
                else:
                    query = query.with_limit(size)
            elif isinstance(part, TailCommand):
                size = HeadCommand.parse_size(arg)
                limit = query.parts[0].limit or Limit(0, size)
                if first_head_tail_in_a_row and head_tail_keep_order:
                    query = query.with_limit(Limit(limit.offset + max(0, limit.length - size), min(limit.length, size)))
                elif first_head_tail_in_a_row and not head_tail_keep_order:
                    query = query.with_limit(Limit(limit.offset, min(limit.length, size)))
                else:
                    head_tail_keep_order = False
                    query = query.with_limit(size)
                    p = query.current_part
                    # the limit might have created a new part - make sure there is a sort order
                    p = p if p.sort else evolve(p, sort=DefaultSort)
                    # reverse the sort order -> limit -> reverse the result
                    query.parts[0] = evolve(p, sort=[s.reversed() for s in p.sort], reverse_result=True)
            else:
                raise AttributeError(f"Do not understand: {part} of type: {class_fqn(part)}")

            # Remember the first head tail in a row of head tails
            if isinstance(part, (HeadCommand, TailCommand)):
                if not first_head_tail_in_a_row:
                    first_head_tail_in_a_row = part
            else:
                first_head_tail_in_a_row = None
                head_tail_keep_order = True

            # Define default sort order, if not already defined
            # A sort order is required to always return the result in a deterministic way to the user.
            # Deterministic order is required for head/tail to work
            parts = [pt if pt.sort else evolve(pt, sort=DefaultSort) for pt in query.parts]
            query = evolve(query, parts=parts)

        # If the last part is a navigation, we need to add sort which will ingest a new part.
        with_sort = query.set_sort(*DefaultSort) if query.current_part.navigation else query
        # When all parts are combined, interpret the result on defined section.
        final_query = with_sort.on_section(ctx.env.get("section", PathRoot))
        options = ExecuteSearchCommand.argument_string(parsed_options)
        query_string = str(final_query)
        execute_search = self.command("execute_search", f"{options}'{query_string}'", ctx)
        return final_query, parsed_options, [execute_search, *additional_commands]

    async def evaluate_cli_command(
        self, cli_input: str, context: CLIContext = EmptyContext, replace_place_holder: bool = True
    ) -> List[ParsedCommandLine]:
        async def combine_query_parts(
            commands: List[ExecutableCommand], ctx: CLIContext
        ) -> Tuple[CLIContext, List[ExecutableCommand]]:
            parts = list(takewhile(lambda x: isinstance(x.command, SearchCLIPart), commands))
            if parts:
                query, options, query_parts = await self.create_query(parts, ctx)
                ctx_wq = evolve(ctx, query=query, query_options=options, commands=commands)
                remaining = executable_commands(
                    commands[len(parts) :], ctx_wq, offset=len(parts), previous_command=query_parts[-1]
                )
                rewritten_parts = [*query_parts, *remaining]
            else:
                ctx_wq = evolve(ctx, commands=commands)
                rewritten_parts = executable_commands(commands, ctx_wq)
            # re-evaluate remaining commands - to take the adapted context into account
            return ctx_wq, rewritten_parts

        def rewrite_command_line(cmds: List[ExecutableCommand], ctx: CLIContext) -> List[ExecutableCommand]:
            """
            Rewrite the command line to make it more user-friendly.
            Rules:
            - add the list command if no output format is defined
            - add a format to write commands if no output format is defined
            - report benchmark run will be formatted as benchmark result automatically
            """
            if ctx.env.get("no_rewrite") or len(cmds) == 0:
                return cmds
            first_cmd = cmds[0]
            last_cmd = cmds[-1]
            single = cmds[0] if len(cmds) == 1 else None
            result = cmds

            def no_format() -> bool:
                return not any(c for c in result if isinstance(c.command, (OutputTransformer, PreserveOutputFormat)))

            def fmt_benchmark() -> ExecutableCommand:
                return self.command("format", "--benchmark-result", ctx)

            def fmt_list() -> ExecutableCommand:
                return self.command("list", None, ctx)

            # benchmark run as single command is rewritten to benchmark run | format --benchmark-result
            if single and isinstance(single.command, ReportCommand) and ReportCommand.is_run_action(single.arg):
                result = [single, fmt_benchmark()]
            # if the last command is a write command without any format: add the format
            elif isinstance(last_cmd.command, WriteCommand) and no_format():
                # format is either list (default) or benchmark
                fmt = fmt_benchmark() if isinstance(first_cmd.command, ReportCommand) else fmt_list()
                result = [*cmds[0:-1], fmt, cmds[-1]]

            # produces text and no resulting output transformer is defined: add the default `list` command
            if last_cmd.action.produces.text and no_format():
                result = [*result, fmt_list()]
            return result

        def adjust_context(parsed: ParsedCommands) -> CLIContext:
            cmd_env = {**self.cli_env, **context.env, **parsed.env}
            ctx = evolve(context, env=cmd_env)
            last_command = self.commands.get(parsed.commands[-1].cmd) if parsed.commands else None
            if isinstance(last_command, NoTerminalOutput) and ctx.console_renderer:
                return evolve(ctx, env=cmd_env, console_renderer=ConsoleRenderer.default_renderer())
            else:
                return evolve(context, env=cmd_env)

        # iterate the list of commands and pass information about position, previous command, etc.
        def executable_commands(
            commands: Sequence[Union[ParsedCommand, ExecutableCommand]],
            ctx: CLIContext,
            *,
            previous_command: Optional[ExecutableCommand] = None,
            offset: int = 0,
        ) -> List[ExecutableCommand]:
            result: List[ExecutableCommand] = []
            for pos, c in enumerate(commands):
                name, arg = (c.cmd, c.args) if isinstance(c, ParsedCommand) else (c.name, c.arg)
                command = self.command(name, arg, ctx, position=pos + offset, previous_command=previous_command)
                previous_command = command
                result.append(command)
            return result

        async def parse_line(parsed: ParsedCommands) -> ParsedCommandLine:
            ctx = adjust_context(parsed)
            executable = executable_commands(parsed.commands, ctx)
            rewritten = rewrite_command_line(executable, ctx)
            ctx, commands = await combine_query_parts(rewritten, ctx)
            not_met = [r for cmd in commands for r in cmd.action.required if r.name not in context.uploaded_files]
            envelope = {k: v for cmd in commands for k, v in cmd.action.envelope.items()}
            return ParsedCommandLine(ctx, parsed, commands, not_met, envelope)

        def expand_aliases(line: ParsedCommands) -> ParsedCommands:
            def expand_alias(alias_cmd: ParsedCommand) -> List[ParsedCommand]:
                alias: AliasTemplate = self.alias_templates[alias_cmd.cmd]
                available: Dict[str, AliasTemplateParameter] = {p.name: p for p in alias.parameters}
                props: Dict[str, JsonElement] = self.replacements(**{**self.cli_env, **context.env})  # type: ignore
                props["args"] = alias_cmd.args or ""
                for p in alias.parameters:
                    props[p.name] = p.default
                # only parse properties, if there are any declared
                if alias.parameters:
                    args = (alias_cmd.args or "").strip()
                    parser = args_values_parser if args.startswith("--") else key_values_parser
                    props.update(parser.parse(args))
                undefined = [
                    available[k].arg_name for k, v in props.items() if k != "args" and v is None and k in available
                ]
                if undefined:
                    raise AttributeError(
                        f"Alias {alias_cmd.cmd} not enough parameters provided. Missing: {', '.join(undefined)}"
                    )
                rendered = alias.render(props)
                log.debug(f"The rendered alias template is: {rendered}")
                return single_commands.parse(rendered)  # type: ignore

            def expand_infra_app_alias(alias_cmd: ParsedCommand) -> List[ParsedCommand]:
                alias: InfraAppAlias = self.infra_app_aliases[alias_cmd.cmd]
                props: Dict[str, JsonElement] = self.replacements(**{**self.cli_env, **context.env})  # type: ignore
                props["args"] = alias_cmd.args or ""
                rendered = alias.render(props)
                log.debug(f"The rendered infra app alias template is: {rendered}")
                return single_commands.parse(rendered)  # type: ignore

            result: List[ParsedCommand] = []
            for cmd in line.commands:
                if cmd.cmd in self.alias_templates:
                    result.extend(expand_alias(cmd))
                elif cmd.cmd in self.infra_app_aliases:
                    result.extend(expand_infra_app_alias(cmd))
                else:
                    result.append(cmd)

            return ParsedCommands(result, line.env)

        async def send_analytics(parsed: List[ParsedCommands], raw: List[ParsedCommands]) -> None:
            command_names = [cmd.cmd for line in parsed for cmd in line.commands]
            used_aliases = [cmd.cmd for line in raw for cmd in line.commands if cmd.cmd in self.alias_templates]
            used_infra_app_aliases = [
                cmd.cmd for line in raw for cmd in line.commands if cmd.cmd in self.infra_app_aliases
            ]
            resoto_session_id = context.env.get("resoto_session_id")
            await self.dependencies.event_sender.core_event(
                CoreEvent.CLICommand,
                {
                    "command_names": command_names,
                    "used_aliases": used_aliases,
                    "used_infra_app_aliases": used_infra_app_aliases,
                    "session_id": resoto_session_id,
                    "source": context.source or "unknown",
                },
                command_lines=len(parsed),
                commands=len(command_names),
            )

        def replace_placeholders(parsed: ParsedCommands) -> ParsedCommands:
            cmd_env = {**self.cli_env, **context.env, **parsed.env}

            def replace_command(cmd: ParsedCommand) -> ParsedCommand:
                args = cmd.args if cmd.args is None else self.replace_placeholder(cmd.args, **cmd_env)
                return ParsedCommand(cmd.cmd, args)

            return ParsedCommands([replace_command(cmd) for cmd in parsed.commands], parsed.env)

        # parse command lines (raw)
        raw_parsed: List[ParsedCommands] = multi_command_parser.parse(cli_input)
        # expand aliases
        command_lines = [expand_aliases(cmd_line) for cmd_line in raw_parsed]
        # send analytics
        await send_analytics(command_lines, raw_parsed)
        # decide, if placeholders should be replaced
        first_command = command_lines[0].commands[0] if command_lines and command_lines[0].commands else None
        keep_raw = not replace_place_holder or (first_command and JobsCommand.is_jobs_update(first_command))
        command_lines = command_lines if keep_raw else [replace_placeholders(cmd_line) for cmd_line in command_lines]
        res = [await parse_line(cmd_line) for cmd_line in command_lines]
        return res

    async def execute_cli_command(self, cli_input: str, sink: Sink[T], ctx: CLIContext = EmptyContext) -> List[T]:
        return [await parsed.to_sink(sink) for parsed in await self.evaluate_cli_command(cli_input, ctx, True)]



import logging
from re import RegexFlag, fullmatch
from typing import Optional, Callable, Awaitable, Tuple

from aiohttp.hdrs import METH_OPTIONS, METH_GET, METH_HEAD, METH_POST, METH_PUT, METH_DELETE, METH_PATCH
from aiohttp.web import HTTPRedirection, HTTPNotFound, HTTPBadRequest, HTTPException, HTTPNoContent
from aiohttp.web_exceptions import HTTPServiceUnavailable, HTTPError
from aiohttp.web_middlewares import middleware
from aiohttp.web_request import Request
from aiohttp.web_response import StreamResponse

from resotocore import version
from resotocore.analytics import AnalyticsEventSender, CoreEvent
from resotocore.core_config import CoreConfig
from resotocore.error import NotFoundError, ClientError
from resotocore.metrics import RequestInProgress, RequestLatency, RequestCount, perf_now
from resotocore.web import RequestHandler, api  # pylint: disable=unused-import # prevent circular import

log = logging.getLogger(__name__)


def enable_compression(request: Request, response: StreamResponse) -> None:
    # The UI can not handle compressed responses. Allow compression only if requested by somebody else
    if "resotoui-via" not in request.headers:
        response.enable_compression()


async def on_response_prepare(request: Request, response: StreamResponse) -> None:
    # Headers are required for the UI to work, since it uses SharedArrayBuffer.
    # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/SharedArrayBuffer
    if fullmatch("/ui/.*", request.path, RegexFlag.IGNORECASE):
        response.headers["Cross-Origin-Opener-Policy"] = "same-origin"
        response.headers["Cross-Origin-Embedder-Policy"] = "require-corp"

    # In case of a CORS request: a response header to allow the origin is required
    if request.headers.get("sec-fetch-mode") == "cors":
        response.headers["Access-Control-Allow-Origin"] = request.headers.get("origin", "*")

    if request.method in (METH_GET, METH_HEAD, METH_POST, METH_PUT, METH_DELETE, METH_PATCH):
        response.headers["Access-Control-Allow-Origin"] = request.headers.get("origin", "*")
        response.headers["Access-Control-Allow-Methods"] = request.headers.get("access-control-request-method", "*")
        response.headers["Access-Control-Allow-Headers"] = request.headers.get("access-control-request-headers", "*")
        response.headers["Access-Control-Max-Age"] = "86400"

    # If requested, send the JWT in the response header
    if auth_header := request.get("send_auth_response_header"):
        response.headers["Authorization"] = auth_header


@middleware
async def cors_handler(request: Request, handler: RequestHandler) -> StreamResponse:
    if request.method == METH_OPTIONS:
        return HTTPNoContent(
            headers={
                # allow origin of request or all if none is defined.
                "Access-Control-Allow-Origin": request.headers.get("origin", "*"),
                # allow the requested method or all if none is defined.
                "Access-Control-Allow-Methods": request.headers.get("access-control-request-method", "*"),
                # allow the requested header names or all if none is defined.
                "Access-Control-Allow-Headers": request.headers.get("access-control-request-headers", "*"),
                # allow the client to cache this result
                "Access-Control-Max-Age": "86400",  # allow caching for one day
            }
        )
    else:
        return await handler(request)


@middleware
async def metrics_handler(request: Request, handler: RequestHandler) -> StreamResponse:
    request["start_time"] = perf_now()
    RequestInProgress.labels(request.path, request.method).inc()
    try:
        response = await handler(request)
        RequestCount.labels(request.method, request.path, response.status).inc()
        return response
    except HTTPException as ex:
        RequestCount.labels(request.method, request.path, ex.status).inc()
        raise ex
    finally:
        resp_time = perf_now() - request["start_time"]
        RequestLatency.labels(request.path).observe(resp_time)
        RequestInProgress.labels(request.path, request.method).dec()


def error_handler(
    config: CoreConfig, event_sender: AnalyticsEventSender
) -> Callable[[Request, RequestHandler], Awaitable[StreamResponse]]:
    is_debug = (logging.root.level < logging.INFO) or config.runtime.debug

    def exc_info(ex: Exception) -> Optional[Exception]:
        return ex if is_debug else None

    @middleware
    async def error_handler_middleware(request: Request, handler: RequestHandler) -> StreamResponse:
        def message_from_error(e: Exception) -> Tuple[str, str, str]:
            kind = type(e).__name__
            ex_str = str(e)
            return f"Error: {kind}\nMessage: {ex_str}", kind, ex_str

        try:
            return await handler(request)
        except HTTPRedirection as e:
            # redirects are implemented as exceptions in aiohttp for whatever reason...
            raise e
        except HTTPError as e:
            log.warning(f"Request {request} has failed with exception: {str(e)}")
            raise e
        except NotFoundError as e:
            message, _, _ = message_from_error(e)
            log.info(f"Request {request} has failed with exception: {message}", exc_info=exc_info(e))
            raise HTTPNotFound(text=message) from e
        except (ClientError, AttributeError) as e:
            message, kind, ex_str = message_from_error(e)
            log.info(f"Request {request} has failed with exception: {message}", exc_info=exc_info(e))
            await event_sender.core_event(
                CoreEvent.ClientError, {"version": version(), "kind": kind, "message": ex_str}
            )
            raise HTTPBadRequest(text=message) from e
        except Exception as e:
            message, kind, ex_str = message_from_error(e)
            log.warning(f"Request {request} has failed with exception: {message}", exc_info=exc_info(e))
            await event_sender.core_event(
                CoreEvent.ServerError, {"version": version(), "kind": kind, "message": ex_str}
            )
            raise HTTPBadRequest(text=message) from e

    return error_handler_middleware


def default_middleware(api_handler: "api.Api") -> Callable[[Request, RequestHandler], Awaitable[StreamResponse]]:
    @middleware
    async def default_handler(request: Request, handler: RequestHandler) -> StreamResponse:
        if api_handler.in_shutdown:
            # We are currently in shutdown: inform the caller to retry later.
            return HTTPServiceUnavailable(headers={"Retry-After": "5"})
        else:
            return await handler(request)

    return default_handler



from typing import AsyncIterator, AsyncGenerator, Optional, List, Union

from aiostream import stream
from aiostream.core import Stream
from networkx import DiGraph
from rich._emoji_codes import EMOJI

from resotocore.model.resolve_in_graph import NodePath
from resotocore.report import BenchmarkResult, CheckCollectionResult, CheckResult, ReportSeverityPriority
from resotocore.types import JsonElement
from resotocore.util import value_in_path

kind_reader = {
    "report_benchmark": BenchmarkResult.from_node,
    "report_check_collection": CheckCollectionResult.from_node,
    "report_check_result": CheckResult.from_node,
}
check_mark = EMOJI["white_check_mark"]
cross_mark = EMOJI["cross_mark"]


def render_benchmark_result(benchmark: BenchmarkResult, account: str) -> str:
    filtered = benchmark.filter_result(failed_for_account=account) if benchmark.only_failed else benchmark
    result = f"# Report for account {account}\n\n"
    result += f"Title: {benchmark.title}\n\n"
    result += f"Version: {benchmark.version}\n\n"
    passing, failing = benchmark.passing_failing_checks_count_for_account(account)
    if passing == 0:
        result += f"Summary: all {failing} checks failed\n\n"
    elif failing == 0:
        result += f"Summary: all {passing} checks passed\n\n"
    else:
        result += f"Summary: {passing} checks {check_mark} and {failing} checks {cross_mark}\n\n"
    # If there are no children, ignore the summary (same information will be available as part of the check result)
    if filtered.children:
        result += render_benchmark_summary(filtered, account)
        result += "\n\n"
    for check in filtered.checks:
        result += render_check_result(check, account)
    for collection in filtered.children:
        result += render_collection_result(collection, account, 2)
    return result


def render_benchmark_summary(benchmark: CheckCollectionResult, account: str) -> str:
    passing, failing = benchmark.passing_failing_checks_for_account(account)

    def render_result_list(name: str, icon: str, checks: List[CheckResult]) -> str:
        if checks:
            lr = f"## {name} \n\n"
            for check in sorted(checks, key=lambda x: -ReportSeverityPriority[x.check.severity]):
                lr += f"- {icon} {check.check.severity.name}: {check.check.title}\n"
            return lr
        else:
            return ""

    result = render_result_list("Passed Checks", check_mark, passing)
    result += render_result_list("Failed Checks", cross_mark, failing)
    return result


def render_collection_result(collection: CheckCollectionResult, account: str, level: int) -> str:
    result = f"{'#' * level} {collection.title} "
    passing, failing = collection.passing_failing_checks_count_for_account(account)
    if passing == 0:
        result += f"(all checks {cross_mark})"
    elif failing == 0:
        result += f"(all checks {check_mark})"
    else:
        result += f"({passing} checks {check_mark}, {failing} checks {cross_mark})"
    result += f"\n\n{collection.description}\n\n"
    for check in collection.checks:
        result += render_check_result(check, account)
    for sub_collection in collection.children:
        result += render_collection_result(sub_collection, account, level + 1)
    return result


def render_check_result(check_result: CheckResult, account: str) -> str:
    check = check_result.check
    failed = account in check_result.number_of_resources_failing_by_account
    result = f"- {cross_mark if failed else check_mark} **{check.severity.name}**: {check.title}\n\n"
    if failed:
        result += f"  - Risk: {check.risk}\n\n"
        result += f"  - There are {check_result.number_of_resources_failing_by_account[account]} `{check.result_kind}` "
        result += "resources failing this check.\n\n"
        result += (
            f"  - Remediation: {check.remediation.text}. See [Link]({check.remediation.url}) for more details.\n\n"
        )
    return result


async def respond_benchmark_result(gen: Union[Stream, AsyncIterator[JsonElement]]) -> AsyncGenerator[str, None]:
    # step 1:  read graph
    graph = DiGraph()
    async with stream.iterate(gen).stream() as streamer:
        async for item in streamer:
            if isinstance(item, dict):
                type_name = item.get("type")
                if type_name == "node":
                    uid = value_in_path(item, NodePath.node_id)
                    reported = value_in_path(item, NodePath.reported)
                    kind = value_in_path(item, NodePath.reported_kind)
                    if uid and reported and kind and (reader := kind_reader.get(kind)):
                        graph.add_node(uid, data=reader(item))
                elif type_name == "edge":
                    from_node = value_in_path(item, NodePath.from_node)
                    to_node = value_in_path(item, NodePath.to_node)
                    if from_node and to_node:
                        graph.add_edge(from_node, to_node)
                else:
                    raise AttributeError(f"Expect json object but got: {type(item)}: {item}")
            else:
                raise AttributeError(f"Expect json object but got: {type(item)}: {item}")

    # step 2: read benchmark result from graph
    def traverse(node_id: str, collection: CheckCollectionResult) -> None:
        for sub_id in graph.successors(node_id):
            node = graph.nodes[sub_id]
            collection_node = node["data"]
            if isinstance(collection_node, CheckCollectionResult):
                collection.children.append(collection_node)
                traverse(sub_id, collection_node)
            elif isinstance(collection_node, CheckResult):
                collection.checks.append(collection_node)

    result: Optional[BenchmarkResult] = None
    for nid, data in graph.nodes(data=True):
        if isinstance(data.get("data"), BenchmarkResult):
            br = data["data"]
            traverse(nid, br)
            result = br

    # step 3: benchmark result to markdown
    if result:
        # use accounts defined on the cmd line, otherwise fall back to accounts with failing checks
        accounts = result.accounts or result.failing_accounts()
        for account in accounts:
            yield render_benchmark_result(result, account)



# key: operation, value: reverse operation
# note: the reverse operation is not the negated operation
# (negation of < would be >=)
from resotocore.model.resolve_in_graph import NodePath

less_greater_then_operations = {
    "<": ">",
    "<=": ">=",
    ">": "<",
    ">=": "<=",
}

# operations that arangodb matches to true, if the value is not existent.
# example foo < 3 would match, if property foo does not exist.
# see: https://www.arangodb.com/docs/stable/aql/fundamentals-type-value-order.html
arangodb_matches_null_ops = {"<", "<=", "not in", "=~"}

# Attribute blacklist to hide in text/plain mode.
plain_text_blacklist = [
    NodePath.type,
    NodePath.revision,
    NodePath.kinds,
    NodePath.descendant_count,
    NodePath.descendant_summary,
    NodePath.python_type,
]



import pytest
from aiostream import stream

from resotocore.report import Benchmark
from resotocore.report.benchmark_renderer import respond_benchmark_result
from resotocore.report.inspector_service import InspectorService
from resotocore.ids import GraphName


@pytest.mark.asyncio
async def test_benchmark_renderer(inspector_service: InspectorService, test_benchmark: Benchmark) -> None:
    bench_result = await inspector_service.perform_benchmark(GraphName("ns"), "test")
    render_result = [elem async for elem in respond_benchmark_result(stream.iterate(bench_result.to_graph()))]
    assert len(render_result) == 1
    assert (
        render_result[0]
        == "# Report for account sub_root\n\nTitle: test\n\nVersion: 1.5\n\nSummary: all 2 checks failed\n\n## Failed Checks \n\n- ❌ medium: Test\n- ❌ medium: Test\n\n\n## Section 1 (all checks ❌)\n\nTest section.\n\n- ❌ **medium**: Test\n\n  - Risk: Some risk\n\n  - There are 11 `foo` resources failing this check.\n\n  - Remediation: Some remediation text. See [Link](https://example.com) for more details.\n\n## Section 2 (all checks ❌)\n\nTest section.\n\n- ❌ **medium**: Test\n\n  - Risk: Some risk\n\n  - There are 11 `foo` resources failing this check.\n\n  - Remediation: Some remediation text. See [Link](https://example.com) for more details.\n\n"  # noqa: E501
    )

    # only render checks
    check_result = bench_result.to_graph(True)
    assert len(check_result) == 2
    for c in check_result:
        assert c["kind"] == "report_check_result"



from typing import List, Optional, Dict

from resotocore.task import TaskHandler, RunningTaskInfo
from resotocore.task.task_description import RunningTask, Job, Workflow
from resotocore.ids import TaskDescriptorId
from resotocore.util import first


class InMemJobHandler(TaskHandler):
    def __init__(self) -> None:
        self.jobs: List[Job] = []
        self.started_tasks: List[TaskDescriptorId] = []

    async def list_jobs(self) -> List[Job]:
        return self.jobs

    async def list_workflows(self) -> List[Workflow]:
        return []

    async def add_job(
        self,
        job: Job,
        force: bool = False,
    ) -> None:
        self.jobs.append(job)

    async def delete_job(self, job_id: str, force: bool = False) -> Optional[Job]:
        job: Optional[Job] = first(lambda j: j.id == job_id, self.jobs)
        if job:
            self.jobs.remove(job)
            return job
        else:
            return None

    async def parse_job_line(
        self, source: str, line: str, env: Optional[Dict[str, str]] = None, mutable: bool = True
    ) -> Job:
        raise NotImplementedError()

    async def start_task_by_descriptor_id(self, uid: TaskDescriptorId) -> Optional[RunningTaskInfo]:
        self.started_tasks.append(uid)
        return None

    async def running_tasks(self) -> List[RunningTask]:
        raise NotImplementedError()



from __future__ import annotations

import logging
from abc import ABC
from asyncio import Queue
from contextlib import asynccontextmanager
from datetime import datetime
from typing import Any, Optional, Dict, List, AsyncGenerator

from frozendict import frozendict
from jsons import set_deserializer, set_serializer
from resotocore.ids import TaskId
from resotocore.service import Service

from resotocore.types import Json
from resotocore.util import pop_keys, utc_str, from_utc
from resotocore.ids import SubscriberId
from resotolib.core.progress import Progress

log = logging.getLogger(__name__)


class CoreMessage:
    Connected = "message-listener-connected"
    Disconnected = "message-listener-disconnected"
    ConfigUpdated = "config-updated"
    ConfigCopied = "config-copied"
    ConfigDeleted = "config-deleted"
    ErrorMessage = "error"
    ProgressMessage = "progress"
    GraphMergeCompleted = "graph-merge-completed"


class Message(ABC):
    """
    Json representation of a message is always:
    { "kind": "xxx", "message_type": "yyy", "data": { ... }}

    The kind defines the purpose of this message and is one of those:
    - event: something happened in the system
    - action: some action needs to be performed
    - action_done: response for a given action to mark this action as done
    - action_error: response for a given action to mark this action as error

    The message type is an identifier that identifies the cause of a message.
    A message_type has to be unique for a specific cause.
    Subscribers subscribe solely on message_type.

    The data field can hold arbitrary data that makes sense for the specific message.
    For all action, action_done and action_error messages, the data field contains references to the task.
    """

    def __init__(self, message_type: str, data: Optional[Json]):
        self.message_type = message_type
        self.data = frozendict(data if data else {})

    def __eq__(self, other: Any) -> bool:
        return self.__dict__ == other.__dict__ if isinstance(other, Message) else False

    def __hash__(self) -> int:
        return hash(self.message_type) + hash(self.data)

    @staticmethod
    def from_json(json: Json, _: type = object, **__: object) -> Message:
        kind = json["kind"]
        message_type = json["message_type"]
        data: Json = json.get("data", {})
        if kind == "event":
            res_data = pop_keys(data, ["subscriber_id"])
            return Event(message_type, res_data)
        elif kind == "action":
            res_data = pop_keys(data, ["task", "step", "subscriber_id"])
            return Action(message_type, data["task"], data["step"], res_data)
        elif kind == "action_done":
            res_data = pop_keys(data, ["task", "step", "subscriber_id"])
            return ActionDone(message_type, data["task"], data["step"], data["subscriber_id"], res_data)
        elif kind == "action_info":
            return ActionInfo(
                message_type, data["task"], data["step"], data["subscriber_id"], data["level"], data["message"]
            )
        elif kind == "action_progress":
            return ActionProgress(
                message_type,
                data["task"],
                data["step"],
                data["subscriber_id"],
                Progress.from_json(data["progress"]),
                from_utc(data["at"]),
            )
        elif kind == "action_error":
            res_data = pop_keys(data, ["task", "step", "subscriber_id", "error"])
            return ActionError(
                message_type, data["task"], data["step"], data["subscriber_id"], data.get("error", "n/a"), res_data
            )
        else:
            raise AttributeError(f"No handler to parse {kind}")

    @staticmethod
    def message_to_json(o: Message, **_: object) -> Json:
        if isinstance(o, Event):
            return {
                "kind": "event",
                "message_type": o.message_type,
                "data": o.data,
            }
        elif isinstance(o, Action):
            extra_data = {"task": o.task_id, "step": o.step_name}
            return {
                "kind": "action",
                "message_type": o.message_type,
                "data": {**o.data, **extra_data},
            }
        elif isinstance(o, ActionDone):
            extra_data = {"task": o.task_id, "step": o.step_name, "subscriber_id": o.subscriber_id}
            return {"kind": "action_done", "message_type": o.message_type, "data": {**o.data, **extra_data}}
        elif isinstance(o, ActionProgress):
            extra_data = {"task": o.task_id, "step": o.step_name, "subscriber_id": o.subscriber_id}
            return {
                "kind": "action_progress",
                "message_type": o.message_type,
                "data": {**o.data, **extra_data, "progress": o.progress.to_json(), "at": utc_str(o.at)},
            }
        elif isinstance(o, ActionInfo):
            extra_data = {"task": o.task_id, "step": o.step_name, "subscriber_id": o.subscriber_id}
            return {
                "kind": "action_info",
                "message_type": o.message_type,
                "data": {**o.data, **extra_data, "level": o.level, "message": o.message},
            }
        elif isinstance(o, ActionError):
            extra_data = {
                "task": o.task_id,
                "step": o.step_name,
                "subscriber_id": o.subscriber_id,
                "error": o.error,
            }
            return {
                "kind": "action_error",
                "message_type": o.message_type,
                "data": {**o.data, **extra_data},
            }
        else:
            raise AttributeError(f"No handler to marshal {type(o).__name__}")


class Event(Message):
    def __init__(self, message_type: str, data: Optional[Json] = None):
        super().__init__(message_type, data)


class ActionMessage(Message):
    def __init__(self, message_type: str, task_id: TaskId, step_name: str, data: Optional[Json] = None):
        super().__init__(message_type, data)
        self.task_id = task_id
        self.step_name = step_name


class Action(ActionMessage):
    def done(self, subscriber_id: SubscriberId) -> ActionDone:
        return ActionDone(self.message_type, self.task_id, self.step_name, subscriber_id, dict(self.data))


class ActionDone(ActionMessage):
    def __init__(
        self,
        message_type: str,
        task_id: TaskId,
        step_name: str,
        subscriber_id: SubscriberId,
        data: Optional[Json] = None,
    ):
        super().__init__(message_type, task_id, step_name, data)
        self.subscriber_id = subscriber_id


class ActionProgress(ActionMessage):
    def __init__(
        self,
        message_type: str,
        task_id: TaskId,
        step_name: str,
        subscriber_id: SubscriberId,
        progress: Progress,
        at: datetime,
    ):
        super().__init__(message_type, task_id, step_name)
        self.subscriber_id = subscriber_id
        self.progress = progress
        self.at = at


class ActionInfo(ActionMessage):
    def __init__(
        self,
        message_type: str,
        task_id: TaskId,
        step_name: str,
        subscriber_id: SubscriberId,
        level: str,
        message: str,
    ):
        super().__init__(message_type, task_id, step_name)
        self.subscriber_id = subscriber_id
        self.level = level
        self.message = message

    def info(self) -> str:
        return f"{self.level.capitalize()}: could not perform action {self.step_name}. Reason: {self.message}"


class ActionError(ActionMessage):
    def __init__(
        self,
        message_type: str,
        task_id: TaskId,
        step_name: str,
        subscriber_id: SubscriberId,
        error: str,
        data: Optional[Json] = None,
    ):
        super().__init__(message_type, task_id, step_name, data)
        self.subscriber_id = subscriber_id
        self.error = error

    def info(self) -> str:
        return f"Fatal: could not perform action {self.step_name}. Reason: {self.error}"


class MessageBus(Service):
    """
    This class implements a simple event bus.
    Every subscriber is context managed and gets its own queue of events.
    """

    def __init__(self) -> None:
        # key is the channel name, value is the list of queues
        self.listeners: Dict[str, List[Queue[Message]]] = {}
        # key is the subscriber id, value is the list of queue names
        self.active_listener: Dict[SubscriberId, List[str]] = {}

    @asynccontextmanager
    async def subscribe(
        self, subscriber_id: SubscriberId, channels: Optional[List[str]] = None, queue_size: int = 0
    ) -> AsyncGenerator[Queue[Message], None]:
        """
        Subscribe to a list of event channels.
        All events that match the channel will be written to this queue.
        This is an async queue - all operations are async!

        Usage Subscriber:
        with bus.subscribe(["foo", "bar", "bla"]) as q:
          elem = await q.get()

        Usage Emitter:
        await bus.emit("foo", { "my": "event" })

        :param subscriber_id: the id of the subscriber.
        :param channels: the list of channels to subscribe to. In case if empty list: all channels.
        :param queue_size: the size of elements that can be buffered in the queue.
        :return: the context managed queue.
        """
        queue: Queue[Message] = Queue(queue_size)

        def add_listener(name: str) -> None:
            if name not in self.listeners:
                self.listeners[name] = [queue]
            else:
                self.listeners[name].append(queue)

        def remove_listener(name: str) -> None:
            self.listeners[name].remove(queue)
            if len(self.listeners[name]) == 0:
                del self.listeners[name]

        ch_list = channels if channels else ["*"]
        if len(ch_list) == 0:
            raise AttributeError("Need at least one channel to subscribe to!")
        try:
            await self.emit_event(CoreMessage.Connected, {"subscriber_id": subscriber_id, "channels": channels})
            self.active_listener[subscriber_id] = ch_list
            for channel in ch_list:
                add_listener(channel)
            log.info(f"Event listener {subscriber_id} added to following queues: {ch_list}")
            yield queue
        finally:
            log.info(f"Remove listener: {subscriber_id}")
            for channel in ch_list:
                remove_listener(channel)
            self.active_listener.pop(subscriber_id, None)
            await self.emit_event(CoreMessage.Disconnected, {"subscriber_id": subscriber_id, "channels": channels})

    async def emit_event(self, event_type: str, data: Json) -> None:
        return await self.emit(Event(event_type, data))

    async def emit(self, message: Message) -> None:
        log.debug(
            "Emitting message %s: %s", message.message_type, ", ".join(f"{k}={v}" for k, v in message.data.items())
        )

        async def emit_by(name: str) -> None:
            for listener in self.listeners.get(name, []):
                await listener.put(message)

        await emit_by(message.message_type)  # inform specific listener
        await emit_by("*")  # inform "all" event listener


set_deserializer(Message.from_json, Message)
set_serializer(Message.message_to_json, Message)



from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import EntityDb, EventEntityDb, ArangoEntityDb
from resotocore.model.model import Kind

ModelDb = EntityDb[str, Kind]
EventModelDb = EventEntityDb[str, Kind]


def model_db(db: AsyncArangoDB, collection: str) -> ArangoEntityDb[str, Kind]:
    return ArangoEntityDb(db, collection, Kind, lambda k: k.fqn)  # type: ignore



import asyncio
import json
from datetime import timedelta
from multiprocessing import set_start_method
from typing import List, AsyncGenerator, Any

import pytest

from resotocore.analytics import AnalyticsEventSender
from resotocore.db.deferred_edge_db import pending_deferred_edge_db
from resotocore.db.graphdb import ArangoGraphDB
from resotocore.db.model import GraphUpdate
from resotocore.system_start import empty_config
from resotocore.ids import TaskId
from resotocore.message_bus import MessageBus
from resotocore.model.db_updater import GraphMerger
from resotocore.model.model import Kind, Model
from resotocore.model.typed_model import to_js
from resotocore.types import Json
from tests.resotocore.db.graphdb_test import create_graph
from tests.resotocore.model import ModelHandlerStatic


@pytest.mark.asyncio
async def test_merge_process(
    event_sender: AnalyticsEventSender, graph_db: ArangoGraphDB, foo_kinds: List[Kind], message_bus: MessageBus
) -> None:
    # set explicitly (is done in main explicitly as well)
    set_start_method("spawn")

    # wipe any existing data
    await graph_db.wipe()
    # store the model in db, so it can be loaded by the sub process
    graph_db.db.collection(f"{graph_db.name}_model").insert_many([to_js(a) for a in foo_kinds])
    # define args to parse for the sub process
    config = empty_config(["--graphdb-username", "test", "--graphdb-password", "test", "--graphdb-database", "test"])
    # create sample graph data to insert
    graph = create_graph("test")

    await pending_deferred_edge_db(graph_db.db, "deferred_outer_edges").create_update_schema()

    async def iterator() -> AsyncGenerator[bytes, None]:
        def to_b(a: Any) -> bytes:
            return bytes(json.dumps(a) + "\n", "utf-8")

        for node in graph.nodes():
            yield to_b(graph.nodes[node])
        for from_node, to_node, data in graph.edges(data=True):
            yield to_b({"from": from_node, "to": to_node, "edge_type": data["edge_type"]})
        yield to_b(
            {"from_selector": {"node_id": "id_123"}, "to_selector": {"node_id": "id_456"}, "edge_type": "delete"}
        )

    model_handler = ModelHandlerStatic(Model.from_kinds(foo_kinds))
    async with GraphMerger(model_handler, event_sender, config, message_bus) as merger:
        result = await merger.merge_graph(
            graph_db, iterator(), timedelta(seconds=30), None, TaskId("test_task_123"), wait_for_result=True
        )
        assert result == GraphUpdate(112, 1, 0, 212, 0, 0)
        elem: Json = graph_db.db.collection("deferred_outer_edges").get("test_task_123")  # type: ignore
        assert elem["_key"] == "test_task_123"
        assert elem["task_id"] == "test_task_123"
        assert elem["edges"][0] == {
            "from_node": {"value": "id_123"},
            "to_node": {"value": "id_456"},
            "edge_type": "delete",
        }

        # make another update without wait
        current_count = graph_db.db.collection("ns").count()
        graph = create_graph("test2", width=1)
        no_result = await merger.merge_graph(
            graph_db, iterator(), timedelta(seconds=30), None, TaskId("test_task_124"), wait_for_result=False
        )
        assert no_result is None

        async def wait_for_result() -> None:
            while graph_db.db.collection("ns").count() == current_count:
                await asyncio.sleep(0.01)

        # make sure that the update is processed in the background
        await asyncio.wait_for(wait_for_result(), timeout=5)



import logging
from abc import ABC, abstractmethod
from typing import Optional, List

from resotolib.durations import DurationRe
from resotocore.model.graph_access import Section
from resotocore.model.model import DateTimeKind
from resotocore.model.resolve_in_graph import NodePath
from resotocore.types import Json
from resotocore.util import value_in_path, from_utc

log = logging.getLogger(__name__)


class AdjustNode(ABC):
    @abstractmethod
    def adjust(self, json: Json) -> Json:
        pass


class NoAdjust(AdjustNode):
    def adjust(self, json: Json) -> Json:
        return json


class DirectAdjuster(AdjustNode):
    # this holds a datetime
    expires_value = [[Section.reported, "tags", "resoto:expires"]]

    # this holds a relative timedelta
    expiration_values = [
        [Section.reported, "tags", "resoto:expiration"],
        [Section.reported, "tags", "expiration"],
    ]

    def adjust(self, json: Json) -> Json:
        def first_matching(paths: List[List[str]]) -> Optional[str]:
            for path in paths:
                value: Optional[str] = value_in_path(json, path)
                if value:
                    return value
            return None

        try:
            expires_tag = first_matching(self.expires_value)
            expires: Optional[str] = None
            if expires_tag:
                expires = DateTimeKind.from_datetime(expires_tag)
            else:
                expiration_tag = first_matching(self.expiration_values)
                if expiration_tag and expires_tag != "never" and DurationRe.fullmatch(expiration_tag):
                    ctime_str = value_in_path(json, NodePath.reported_ctime)
                    if ctime_str:
                        ctime = from_utc(ctime_str)
                        expires = DateTimeKind.from_duration(expiration_tag, ctime)

            if expires:
                if "metadata" not in json:
                    json["metadata"] = {}
                json["metadata"]["expires"] = expires
        except Exception as ex:
            log.debug(f"Could not parse expiration: {ex}")
        # json is mutated to save memory
        return json



from datetime import timedelta

from arango.client import ArangoClient
from arango.database import StandardDatabase

from resotocore.analytics import NoEventSender
from resotocore.db.db_access import DbAccess
from resotocore.system_start import parse_args, empty_config
from resotocore.model.adjust_node import NoAdjust
from resotocore.ids import GraphName

import pytest


def test_already_existing(test_db: StandardDatabase) -> None:
    access = DbAccess(test_db, NoEventSender(), NoAdjust(), empty_config())

    # test db and user already exist
    testdb = ["--graphdb-username", "test", "--graphdb-password", "test", "--graphdb-database", "test"]
    access.connect(parse_args(testdb), timedelta(seconds=0.01), sleep_time=0.01)


def test_not_existing(system_db: StandardDatabase, test_db: StandardDatabase) -> None:
    access = DbAccess(test_db, NoEventSender(), NoAdjust(), empty_config())

    # foo db and pass does not exist
    foodb = ["--graphdb-username", "foo", "--graphdb-password", "test", "--graphdb-database", "foo"]
    system_db.delete_user("foo", ignore_missing=True)
    system_db.delete_database("foo", ignore_missing=True)
    access.connect(parse_args(foodb), timedelta(seconds=5), sleep_time=0.1)
    assert system_db.has_user("foo")
    assert system_db.has_database("foo")


def test_not_existing_and_default_root_account(
    local_client: ArangoClient, system_db: StandardDatabase, test_db: StandardDatabase
) -> None:
    access = DbAccess(test_db, NoEventSender(), NoAdjust(), empty_config())
    # foo db and pass does not exist
    foodb = ["--graphdb-username", "foo", "--graphdb-password", "bombproof", "--graphdb-database", "foo"]
    system_db.delete_user("foo", ignore_missing=True)
    system_db.delete_database("foo", ignore_missing=True)
    access.connect(parse_args(foodb), timedelta(seconds=5), sleep_time=0.1)

    # The default root account is used and a valid password is given -> also the root account uses this password
    changed_root = local_client.db(username="root", password="bombproof")
    # Rest the password to the default one, to reset the state before the test
    changed_root.replace_user("root", "", True)


@pytest.mark.asyncio
async def test_delete_graph(test_db: StandardDatabase) -> None:
    db_access = DbAccess(test_db, NoEventSender(), NoAdjust(), empty_config())
    graph_name = GraphName("test_graph_delete")
    await db_access.create_graph(
        graph_name,
        validate_name=False,
    )
    assert db_access.graph_dbs.get(graph_name)
    await db_access.delete_graph(graph_name)
    assert db_access.graph_dbs.get(graph_name) is None
